import cv2 as cv
from numpy import append
from . import DetectedArea
from .DetectedArea import DetectedFace
from . import ImageManager
from . import Point
from .HelperFunctions import *
import time
import threading
import queue

"""
The idea is that VideoManagers take care of both displaying the original frames when its time and do detections on the given frames.
VideoManager splits the two tasks into two processes/threads?. One thread handles drawing frames and detectedFaces saved in VideoManager.faces array.
The other thread handles running face detection on every other frame/third frame/fourth frame/ however it sees fit-th frame and update the faces location in VideoManager.faces array.

This works by the assumption that for a standard 30fps videos, a normal human being wont be able to move to fast, making the exact location of their face in this frame can be used to 
approximate the upcoming locations in the next few frames. So 
"""


class VideoManager:
    """
    A manager that stores multiple frames (images) and manage how to display them and perform facial detection on those frames.
    """

    HARDCODED_pairOfEyesDistanceRange = (1.5, 3.5)
    HARDCODED_distanceVariation_EyeToCorners_MinOfMax = 0.8
    HARDCODED_updateFaceLocationSearchMultiplier = 1.4
    HARDCODED_similarSizeScale = 0.5
    HARDCODED_faceNotFoundCountLimit = 3

    def __init__(self, videoFps=0, dir=""):
        """
        Construct a VideoManager object
            :param videoFps: the fps of the input video
            :param dir: the directory to the video
        """
        self.videoFps = videoFps
        self.video = cv.VideoCapture(dir)
        self.faces = []

    def displayNonInterferedMethod(self):
        """
        SHOW THE NORMAL JUST GRAB NEWFRAME WHENEVER IT COMES INSTEAD OF LIMITING FRAMERATE THING
        """
        haar_cascasde_face = cv.CascadeClassifier(
            "classifier/haarcascade_frontalface_default.xml")

        frameTimeInterval = 1/self.videoFps
        # if frame is read correctly/ new frame is available, then ret is True.

        prev_frame = time.time()
        while True:
            # Capture frame-by-frame
            ret, frame = self.video.read()
            # if frame is read correctly ret is True
            if not ret:
                print("Can't receive frame (stream end?). Exiting ...")
                break
            # Our operations on the frame come here
            start = time.time()
            frame = resizeMinTo500(frame)
            print(f"DEBUG resize time {time.time() - start}")

            # Display the resulting frame
            next_frame = time.time()
            frameTime = next_frame - prev_frame
            fps = 1/frameTime
            videoSpeed = frameTimeInterval * 100 / frameTime

            cv.putText(frame, f"DEBUG {fps:0.2f} FPS", (12, 20),
                       cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
            cv.putText(frame, f"DEBUG FrameTime {frameTime*1000:0.2f}ms",
                       (12, 60), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
            cv.putText(frame, f"DEBUG video speed {videoSpeed:0.2f}%", (
                12, 100), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
            prev_frame = next_frame

            cv.imshow('60fps maybe?', frame)

            if cv.waitKey(1) == ord('q'):
                break
            # When everything done, release the capture
        self.video.release()
        cv.destroyAllWindows()

    def displayStandardMethod(self):
        """
        SHOW THEM THE STANDARD METHOD PERFORMANCE
        """
        haar_cascasde_face = cv.CascadeClassifier(
            "classifier/haarcascade_frontalface_default.xml")

        # frameTimeInterval, prev_frame, and next_frame are used to determine the frame rate of the video, how often do we check for new frames
        frameTimeInterval = 1/self.videoFps
        prev_frame = time.time()
        next_frame = prev_frame + frameTimeInterval

        # actual_prev_frame and actual_next_frame are for determining the actual fps/frame time when displayed. The deviation from the actual fps/frametime values is because
        # of rendering and image processings applied to the image
        actual_prev_frame = prev_frame
        actual_next_frame = 0

        bool_DisplayingVideo = True

        # startTime, endTime, frameCount are used for debugging (checking average fps)
        debug_startTime = time.time()
        debug_frameCount = 0

        while bool_DisplayingVideo:

            # if its next time for the next frame to be drawn
            if time.time() >= next_frame:

                # Capture frame-by-frame
                ret, frame = self.video.read()

                # if frame is read correctly/ new frame is available, then ret is True.
                if not ret:
                    print("Can't receive frame. Exiting ...")
                    bool_DisplayingVideo = False
                    continue

                # set new frame target (next_frame) after check if new frame is available instead of before. This is to make sure when live capturing
                # the newest image will be shown as soon as they are ready. If this was to be put before checking if new frame is available, newest
                # frame have to wait till the next frameTimeInterval to be shown, which increases latency
                prev_frame = next_frame
                next_frame = prev_frame + frameTimeInterval

                # Our operations on the frame come here
                debug_frameCount = debug_frameCount + 1

                frame = resizeMinTo500(frame)

                # startTime = time.time()
                grayFrame = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
                # print(f"grayscale runtime {time.time() - startTime:0.6f} seconds")

                # startTime = time.time()
                faces = haar_cascasde_face.detectMultiScale(
                    grayFrame, 2, 6, minSize=(30, 30))
                # print(f"Haarcascade Runtime {time.time() - startTime:0.6f} seconds")

                for (x, y, w, h) in faces:
                    cv.rectangle(frame, (x, y), (x+w, y+h), (255, 255, 0), 2)

                # Calculating fps, timeFrame, video speed
                actual_next_frame = time.time()
                frameTime = actual_next_frame - actual_prev_frame

                actual_prev_frame = actual_next_frame
                fps = 1/frameTime
                videoSpeed = frameTimeInterval * 100 / frameTime

                # Display the resulting frame
                cv.putText(frame, f"DEBUG {fps:0.2f} FPS", (12, 20),
                           cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.putText(frame, f"DEBUG FrameTime {frameTime*1000:0.2f}ms",
                           (12, 60), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.putText(frame, f"DEBUG video speed {videoSpeed:0.2f}%", (
                    12, 100), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                # print(f"DEBUG interval {frameTimeInterval*1000:0.2f} frameTime {frameTime*1000:0.2f}")
                cv.imshow('video standard method', frame)

                if cv.waitKey(1) & 0xFF == ord('q'):
                    bool_DisplayingVideo = False
                    continue

        print(f"DEBUG Video length {time.time() - debug_startTime:0.2f}")
        print(
            f"DEBUG Average fps {debug_frameCount / (time.time() - debug_startTime):0.2f} ")

    def displayHowBadUnoptimizedProcessIs(self):
        """
        SHOW THEM HOW BAD IT IS TO JUST RUN DETECTION ON EVERY FRAMES
        """

        # frameTimeInterval, prev_frame, and next_frame are used to determine the frame rate of the video, how often do we check for new frames
        frameTimeInterval = 1/self.videoFps
        prev_frame = time.time()
        next_frame = prev_frame + frameTimeInterval

        # actual_prev_frame and actual_next_frame are for determining the actual fps/frame time when displayed. The deviation from the actual fps/frametime values is because
        # of rendering and image processings applied to the image
        actual_prev_frame = prev_frame
        actual_next_frame = 0

        bool_DisplayingVideo = True

        # startTime, endTime, frameCount are used for debugging (checking average fps)
        debug_startTime = time.time()
        debug_frameCount = 0

        while bool_DisplayingVideo:

            # if its next time for the next frame to be drawn
            if time.time() >= next_frame:

                # Capture frame-by-frame
                ret, frame = self.video.read()

                # if frame is read correctly/ new frame is available, then ret is True.
                if not ret:
                    print("Can't receive frame. Exiting ...")
                    bool_DisplayingVideo = False
                    continue

                # set new frame target (next_frame) after check if new frame is available instead of before. This is to make sure when live capturing
                # the newest image will be shown as soon as they are ready. If this was to be put before checking if new frame is available, newest
                # frame have to wait till the next frameTimeInterval to be shown, which increases latency
                prev_frame = next_frame
                next_frame = prev_frame + frameTimeInterval

                # Our operations on the frame come here
                debug_frameCount = debug_frameCount + 1

                frame = resizeMinTo500(frame)
                frameMgnr = ImageManager(frame)
                # startTime = time.time()
                faces = frameMgnr.findFacesCounterClockwiseMultipleAngles(
                    (0, 45), 1.2, 10)
                # print(f"Runtime {time.time() - startTime:0.6f} seconds")
                for face in faces:
                    face.draw(frame, (255, 255, 255), 2)

                # Calculating fps, timeFrame, video speed
                actual_next_frame = time.time()
                frameTime = actual_next_frame - actual_prev_frame

                actual_prev_frame = actual_next_frame
                fps = 1/frameTime
                videoSpeed = frameTimeInterval * 100 / frameTime

                # Display the resulting frame
                cv.putText(frame, f"DEBUG {fps:0.2f} FPS", (12, 20),
                           cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.putText(frame, f"DEBUG FrameTime {frameTime*1000:0.2f}ms",
                           (12, 60), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.putText(frame, f"DEBUG video speed {videoSpeed:0.2f}%", (
                    12, 100), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.imshow('video unoptimized', frame)

                if cv.waitKey(1) & 0xFF == ord('q'):
                    bool_DisplayingVideo = False
                    continue

        print(f"DEBUG Video length {time.time() - debug_startTime:0.2f}")
        print(
            f"DEBUG Average fps {debug_frameCount / (time.time() - debug_startTime):0.2f} ")

    def displayOptimizedMethod(self):
        """
        SHOW THEM THE "SOMEWHAT" BEST OF YOU
        This method uses the facial detection pipeline written in the ImageManager class (look for eyes, find the pairs, crop out potential faces to run haarcascade facial detection on that area)
        By nature, this method is much more computationally heavier than just running the haarcascade facial detection directly on the image. So instead of running it on every frame, this method only 
        run once its previous call has finished running. This will result in choppy face detection framerate (8 -> 15 fps). 
        """
        # faces is an array that store the most recent position of the detected faces
        faces = []
        # outputQueue is used for storing threaded functions output
        outputQueue = queue.Queue()
        # this is a thread
        faceDetectionThread = threading.Thread()
        faceDetectionThread.start()

        # frameTimeInterval, prev_frame, and next_frame are used to determine the frame rate of the video, how often do we check for new frames
        frameTimeInterval = 1/self.videoFps
        prev_frame = time.time()
        next_frame = prev_frame + frameTimeInterval

        # actual_prev_frame and actual_next_frame are for determining the actual fps/frame time when displayed. The deviation from the actual fps/frametime values is because
        # of rendering and image processings applied to the image
        actual_prev_frame = prev_frame
        actual_next_frame = 0

        bool_DisplayingVideo = True

        # startTime, endTime, frameCount are used for debugging (checking average fps)
        debug_startTime = time.time()
        debug_frameCount = 0
        debug_frameSkip = 0

        while bool_DisplayingVideo:

            # if its next time for the next frame to be drawn
            if time.time() >= next_frame:

                # Capture frame-by-frame
                ret, frame = self.video.read()

                # if frame is read correctly/ new frame is available, then ret is True.
                if not ret:
                    print("Can't receive frame. Exiting ...")
                    bool_DisplayingVideo = False
                    continue

                # resizing the frame to fit my tiny monitor, and to increase haarcascade detection performance too ye
                frame = resizeMinTo500(frame)

                # set new display time target (next_frame) after check if new frame is available instead of before. This is to make sure when live capturing
                # the newest image will be shown as soon as they are ready (variable fps maybe). If this was to be put before checking if new frame is available, newest
                # frame have to wait till the next frameTimeInterval to be shown if it misses the frame display time target, which increases latency (MAYBE :thonking:?)
                prev_frame = next_frame
                next_frame = prev_frame + frameTimeInterval

                # Our operations on the frame come here
                debug_frameCount = debug_frameCount + 1

                # variable face detection scanning fps
                if not faceDetectionThread.is_alive():
                    print(f"DEBUG frameSkip {debug_frameSkip + 1} frames")
                    debug_frameSkip = 0
                    # getting the most recent faces' positions
                    if not outputQueue.empty():
                        faces = outputQueue.get()

                    frameMgnr = ImageManager(frame)
                    args = [(45, 0, -45), 1.2, 10]
                    faceDetectionThread = threading.Thread(
                        target=self.HELPER_findFacesCounterClockwiseMultipleAngles_to_outputQueue, args=[frameMgnr, args, outputQueue])
                    faceDetectionThread.start()
                else:
                    debug_frameSkip = debug_frameSkip + 1

                for face in faces:
                    face.draw(frame, (255, 255, 255), 2)

                # Calculating fps, timeFrame, video speed
                actual_next_frame = time.time()
                frameTime = actual_next_frame - actual_prev_frame

                actual_prev_frame = actual_next_frame
                fps = 1/frameTime
                videoSpeed = frameTimeInterval * 100 / frameTime

                # Display the resulting frame
                cv.putText(frame, f"DEBUG {fps:0.2f} FPS", (12, 20),
                           cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.putText(frame, f"DEBUG FrameTime {frameTime*1000:0.2f}ms",
                           (12, 60), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.putText(frame, f"DEBUG video speed {videoSpeed:0.2f}%", (
                    12, 100), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.imshow('video optimized', frame)

                if cv.waitKey(1) & 0xFF == ord('q'):
                    bool_DisplayingVideo = False
                    continue

        print(f"DEBUG Video length {time.time() - debug_startTime:0.2f}")
        print(
            f"DEBUG Average fps {debug_frameCount / (time.time() - debug_startTime):0.2f} ")

    def displayEvenMoreOptimizedMethod(self):
        """
        SHOW THEM THE BEST OF YOU
        This method uses the facial detection pipeline (macroFacialDetection) written in the ImageManager class (look for eyes, find the pairs, crop out potential faces to run haarcascade facial detection on that area)
        By nature, this method is much more computationally heavier than just running the haarcascade facial detection directly on the image. So instead of running it on every frame, this method only 
        run once its previous call has finished running. This will result in choppy face detection framerate (8 -> 15 fps). 

        To improve the choppy framerate, I use past locations and run basic haarcascade facial detection (microFacialDetection) on those areas only. Then run eye detection in the new detected faces to update the angles of the faces.
        """
        # currentDetectedFacesManager is an array that store the tuples of most recent position of the detected faces and an int indicating to whether should the face be displayed on the frame (0 means yes, > 0 means no)
        currentDetectedFacesManager = []
        # while shadowDetectedFaces store the faces detected by macro facial detection, which lags 2-6 frames behind
        shadowDetectedFaces = []
        # outputQueue is used for storing threaded functions output
        outputQueue = queue.Queue()
        # this is a thread
        faceDetectionThread = threading.Thread()
        faceDetectionThread.start()

        # frameTimeInterval, prev_frame, and next_frame are used to determine the frame rate of the video, how often do we check for new frames
        frameTimeInterval = 1/self.videoFps
        prev_frame = time.time()
        next_frame = prev_frame + frameTimeInterval

        # actual_prev_frame and actual_next_frame are for determining the actual fps/frame time when displayed. The deviation from the actual fps/frametime values is because
        # of rendering and image processings applied to the image
        actual_prev_frame = prev_frame
        actual_next_frame = 0

        bool_DisplayingVideo = True

        # startTime, endTime, frameCount are used for debugging (checking average fps)
        debug_startTime = time.time()
        debug_frameCount = 0
        debug_frameSkip = 0

        while bool_DisplayingVideo:

            # print(f"DEBUG currentDetectedFacesManager length {len(currentDetectedFacesManager)}")
            # print(f"DEBUG shadowDetectedFaces length {len(shadowDetectedFaces)}")

            # if its next time for the next frame to be drawn
            if time.time() >= next_frame:

                # Capture frame-by-frame
                ret, frame = self.video.read()

                # if frame is read correctly/ new frame is available, then ret is True.
                if not ret:
                    print("Can't receive frame. Exiting ...")
                    bool_DisplayingVideo = False
                    continue
                
                startTime = time.time()
                # resizing the frame to fit my tiny monitor, and to increase haarcascade detection performance too ye
                min500frame = resizeMinTo500(frame)
                resizeTime = time.time() - startTime

                # set new display time target (next_frame) after check if new frame is available instead of before. This is to make sure when live capturing
                # the newest image will be shown as soon as they are ready (variable fps maybe). If this was to be put before checking if new frame is available, newest
                # frame have to wait till the next frameTimeInterval to be shown if it misses the frame display time target, which increases latency (MAYBE :thonking:?)
                prev_frame = next_frame
                next_frame = prev_frame + frameTimeInterval

                # Our operations on the frame come here
                debug_frameCount = debug_frameCount + 1

                # variable face detection scanning fps
                if not faceDetectionThread.is_alive():
                    # --- MACRO FACIAL DETECTION SECTION --- MACRO FACIAL DETECTION SECTION --- MACRO FACIAL DETECTION SECTION --- MACRO FACIAL DETECTION SECTION --- MACRO FACIAL DETECTION SECTION --- #

                    # If the previous thoroughly facial detection (my implementation) has finished running
                    # print(f"DEBUG frameSkip {debug_frameSkip + 1} frames")
                    debug_frameSkip = 0

                    # delayDetectedFaces is the array that store detectedArea objs returned by previously called findFacesCounterClockwiseMultipleAngles function.
                    # These objs dont represent the current faces, but faces found in 2-6 frames previously when the function was called
                    delayDetectedFaces = []
                    # getting the most recent thoroughly detected faces' positions
                    if not outputQueue.empty():
                        delayDetectedFaces = outputQueue.get()

                        # print(f"DEBUG delayDetectedFaces length {len(delayDetectedFaces)}")

                    # Merging the macro facial detection into the micro facial detection list
                    if len(shadowDetectedFaces) == 0:
                        shadowDetectedFaces = delayDetectedFaces
                        for shadowFace in shadowDetectedFaces:
                            currentDetectedFacesManager.append(
                                [shadowFace.copy(), 0])
                    elif len(delayDetectedFaces) != 0:

                        # check if there's a new face detected by macro facial detection
                        for shadowFace in shadowDetectedFaces:

                            bool_innerLoop = False
                            i = 0
                            if i < len(delayDetectedFaces):
                                bool_innerLoop = True

                            while bool_innerLoop:
                                delayFace = delayDetectedFaces[i]
                                # if the face returned by macro facial detection has already exists then pop it off
                                if shadowFace.similarSize(delayFace, self.HARDCODED_similarSizeScale) and shadowFace.overlap(delayFace):
                                    delayDetectedFaces.pop(i)
                                    i = i - 1
                                i = i + 1
                                if i >= len(delayDetectedFaces):
                                    bool_innerLoop = False

                        # if there are faces that haven't been detected then add them to the currentDetectedFacesManager list
                        if len(delayDetectedFaces) != 0:
                            for newFace in delayDetectedFaces:
                                currentDetectedFacesManager.append(
                                    [newFace, 0])

                    # Run the next macro facial detection (my implementation) in a seperate thread to avoid interfering with updating new frames
                    frameMgnr = ImageManager(min500frame)
                    args = [(45, 0, -45), 1.1, 10]
                    faceDetectionThread = threading.Thread(
                        target=self.HELPER_findFacesCounterClockwiseMultipleAngles_to_outputQueue, args=[frameMgnr, args, outputQueue])
                    faceDetectionThread.start()

                    # update the shadowDetectedFace list for when the thread finished running
                    for currentFaceManager in currentDetectedFacesManager:
                        currentFace = currentFaceManager[0]
                        shadowDetectedFaces.append(currentFace.copy())

                else:

                    debug_frameSkip = debug_frameSkip + 1

                    # --- MICRO FACIAL DETECTION SECTION --- MICRO FACIAL DETECTION SECTION --- MICRO FACIAL DETECTION SECTION --- MICRO FACIAL DETECTION SECTION --- MICRO FACIAL DETECTION SECTION --- #

                    # updating the detected faces in the last frame by searching around their area
                    if len(currentDetectedFacesManager) != 0:
                        bool_loop = True
                        index_currentDetectedFacesManager = 0
                        if index_currentDetectedFacesManager >= len(currentDetectedFacesManager):
                            bool_loop = False
                        while bool_loop:
                            face = currentDetectedFacesManager[index_currentDetectedFacesManager][0]

                            dimensions = face.dimensions
                            counterClockwiseAngle = face.counterClockwiseAngle
                            faceCenter = Point(face.center.x, face.center.y)

                            # rotate clockwise because the face's angle is the counter clockwise angle of the right eye to the left eye
                            frameOrigin = Point(
                                min500frame.shape[1]/2, min500frame.shape[0]/2)
                            rotatedFrame = rotateClockwise(
                                min500frame, counterClockwiseAngle)
                            rotatedFrameOrigin = Point(
                                rotatedFrame.shape[1]/2, rotatedFrame.shape[0]/2)

                            # rotate clockwise the faceCenter and project it to the new image
                            rotatedFaceCenter = faceCenter.rotatePointClockwise(
                                frameOrigin, counterClockwiseAngle)
                            rotatedFaceCenter = rotatedFaceCenter.projectPoint(
                                frameOrigin, rotatedFrameOrigin)

                            # Calculate the region of the potential face
                            searchingDimensions = (dimensions[0] * self.HARDCODED_updateFaceLocationSearchMultiplier,
                                                   dimensions[1] * self.HARDCODED_updateFaceLocationSearchMultiplier)
                            cropRangeMinY = int(
                                max(0, rotatedFaceCenter.y - searchingDimensions[1] / 2))
                            cropRangeMaxY = int(
                                min(rotatedFrame.shape[0], rotatedFaceCenter.y + searchingDimensions[1] / 2))

                            cropRangeMinX = int(
                                max(0, rotatedFaceCenter.x - searchingDimensions[0] / 2))
                            cropRangeMaxX = int(
                                min(rotatedFrame.shape[1], rotatedFaceCenter.x + searchingDimensions[0] / 2))

                            # Cropping out the region of the potential face
                            try:
                                croppedPotentialFace = np.zeros(
                                    (cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX), dtype='uint8')
                                croppedPotentialFace[0: croppedPotentialFace.shape[0], 0: croppedPotentialFace.shape[1]] = \
                                    rotatedFrame[cropRangeMinY:cropRangeMaxY,
                                                 cropRangeMinX:cropRangeMaxX]

                            except:
                                croppedPotentialFace = np.zeros(
                                    (cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX, 3), dtype='uint8')
                                croppedPotentialFace[0: croppedPotentialFace.shape[0], 0: croppedPotentialFace.shape[1]] = \
                                    rotatedFrame[cropRangeMinY:cropRangeMaxY,
                                                 cropRangeMinX:cropRangeMaxX]

                            croppedPotentialFaceCenter = Point(
                                croppedPotentialFace.shape[1]/2, croppedPotentialFace.shape[0]/2)

                            # gray scale the image
                            grayCroppedPotentialFace = cv.cvtColor(
                                croppedPotentialFace, cv.COLOR_BGR2GRAY)
                            # run basic haarcascade facial detection on it
                            # CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ? CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ? CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ?
                            detectedFaces = haarcascade_FacialDetection(
                                grayCroppedPotentialFace, 1.3, 6)

                            # if found nothing then move to the next face's last frame location
                            if len(detectedFaces) == 0:

                                # increment the faceNotFoundCount
                                faceNotFoundCount = currentDetectedFacesManager[
                                    index_currentDetectedFacesManager][1]
                                faceNotFoundCount = faceNotFoundCount + 1
                                if faceNotFoundCount >= self.HARDCODED_faceNotFoundCountLimit:

                                    currentDetectedFacesManager.pop(
                                        index_currentDetectedFacesManager)
                                    # if there's still more currentFaceManager to iterate
                                    if index_currentDetectedFacesManager < len(currentDetectedFacesManager):
                                        continue

                                    # if there's no more currentFaceManager to iterate then break the loop
                                    bool_loop = False
                                    continue

                                # update faceNotFoundCount
                                currentDetectedFacesManager[index_currentDetectedFacesManager][1] = faceNotFoundCount

                                # if still under the faceNotFoundLimit then continue iterating
                                continue

                            # if found anything, then convert the rectangle coordinates received from detectMultiScale function to a detectedFace obj.
                            biggestFace = detectedFaces[0]
                            for i in range(1, len(detectedFaces)):
                                # comparing the dimensions (width only, cause if width is > then height is also >) among all the faces found.
                                if biggestFace[2] < detectedFaces[i][2]:
                                    biggestFace = detectedFaces[i]

                            # Creating DetectedFace obj representing the biggest face found in the potential region above
                            biggestFace = DetectedFace(
                                (biggestFace[0], biggestFace[1]), (biggestFace[2], biggestFace[3]))

                            # Finding biggestFace's angle
                            # CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ? CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ? CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ?

                            # find the eyes
                            tempEyesArray = haarcascade_EyeDetection(
                                grayCroppedPotentialFace, 1.3, 10)
                            detectedEyes = []

                            if face.leftEye != None:
                                referenceEye = face.leftEye
                            elif face.rightEye != None:
                                referenceEye = face.rightEye
                            else:
                                faceUpperLeftPoint = biggestFace.upperLeft
                                # if the face has no eyes found, then create a standard reference eye
                                # HARDCODED ___ Eye's width is usually at most 3/8 of face's width. Eye's height is usually at most 1/3 of face's height
                                referenceEye = DetectedArea((faceUpperLeftPoint.x + biggestFace.dimensions[0]/8, faceUpperLeftPoint.y + biggestFace.dimensions[1]/6), (
                                    biggestFace.dimensions[0] * 3 / 8, biggestFace.dimensions[1]/3))

                            for (x, y, w, h) in tempEyesArray:
                                eye = DetectedArea((x, y), (w, h))
                                # If the shape and size of the eye found in the image is approximately the same as the reference eye
                                if eye.similarSize(referenceEye, self.HARDCODED_similarSizeScale):
                                    detectedEyes.append(eye)

                            # find the pairs of eyes
                            if len(detectedEyes) > 1:
                                # This might be slightlyyyyy overkill but usually there's like at most 2-3 eyes so it wont be too computationally heavy.
                                for i in range(len(detectedEyes)):
                                    j = i + 1
                                    for j in range(len(detectedEyes)):
                                        if detectedEyes[i].appropriateDistanceTo(detectedEyes[j], self.HARDCODED_pairOfEyesDistanceRange[0], self.HARDCODED_pairOfEyesDistanceRange[1]):
                                            leftEye = detectedEyes[i]
                                            rightEye = detectedEyes[j]
                                            if detectedEyes[i].center.x > detectedEyes[j].center.x:
                                                leftEye = detectedEyes[j]
                                                rightEye = detectedEyes[i]
                                            # if the distance of LeftEye's center to face's upperLeft Point ~ that of RightEye's center to face's upperRight Point
                                            # and distance of LeftEye's center to face's lowerLeft Point ~ that of RightEye's center to face's lowerRight Point
                                            # then this is probably the right pair of eyes
                                            upperLeftDist = leftEye.center.distTo(
                                                biggestFace.upperLeft)
                                            upperRightDist = rightEye.center.distTo(
                                                biggestFace.upperRight)
                                            lowerLeftDist = leftEye.center.distTo(
                                                biggestFace.lowerLeft)
                                            lowerRightDist = rightEye.center.distTo(
                                                biggestFace.lowerRight)
                                            criterias = 0
                                            if upperLeftDist < upperRightDist and upperLeftDist > upperRightDist * self.HARDCODED_distanceVariation_EyeToCorners_MinOfMax:
                                                criterias = criterias + 1
                                            if upperRightDist <= upperLeftDist and upperRightDist > upperLeftDist * self.HARDCODED_distanceVariation_EyeToCorners_MinOfMax:
                                                criterias = criterias + 1
                                            if lowerLeftDist < lowerRightDist and lowerLeftDist > lowerRightDist * self.HARDCODED_distanceVariation_EyeToCorners_MinOfMax:
                                                criterias = criterias + 1
                                            if lowerRightDist <= lowerLeftDist and lowerRightDist > lowerLeftDist * self.HARDCODED_distanceVariation_EyeToCorners_MinOfMax:
                                                criterias = criterias + 1
                                            if criterias == 2:
                                                biggestFace.leftEye = leftEye
                                                biggestFace.rightEye = rightEye
                                                break

                            # find the face's angle using the pair of eyes
                            if biggestFace.leftEye != None and biggestFace.rightEye != None:
                                # calculate teh relative angle created by the two eyes
                                relativeCounterClockwiseAngle = biggestFace.leftEye.center.relativeCounterClockwiseAngle(
                                    biggestFace.rightEye.center)
                                # the face's current angle is the already rotated angle pre-haarcascade detection + the relative angle
                                biggestFace.counterClockwiseAngle = counterClockwiseAngle + \
                                    relativeCounterClockwiseAngle
                                biggestFace.rotateAreaCounterClockwise(
                                    biggestFace.center, relativeCounterClockwiseAngle)
                            else:
                                # if there's not a pair of eyes then the face's angle = the previous face that was detected in the same area's angle
                                biggestFace.counterClockwiseAngle = counterClockwiseAngle

                            # Returning the face's coordinates to the original frame
                            biggestFace.projectArea(
                                croppedPotentialFaceCenter, rotatedFaceCenter)
                            biggestFace.rotateAreaCounterClockwise(
                                rotatedFrameOrigin, biggestFace.counterClockwiseAngle)
                            biggestFace.projectArea(
                                rotatedFrameOrigin, frameOrigin)

                            # update the face's location and reset faceNotFoundCount value to 0
                            currentDetectedFacesManager[index_currentDetectedFacesManager][0] = biggestFace
                            currentDetectedFacesManager[index_currentDetectedFacesManager][1] = 0

                            # continue iterating the currentFaceManager
                            index_currentDetectedFacesManager = index_currentDetectedFacesManager + 1
                            if index_currentDetectedFacesManager >= len(currentDetectedFacesManager):
                                bool_loop = False

                for faceManager in currentDetectedFacesManager:
                    face = faceManager[0]
                    face.draw(min500frame, (255, 255, 255), 2)

                # Calculating fps, timeFrame, video speed
                actual_next_frame = time.time()
                frameTime = actual_next_frame - actual_prev_frame

                actual_prev_frame = actual_next_frame
                fps = 1/frameTime
                videoSpeed = frameTimeInterval * 100 / frameTime

                # Display the resulting frame
                cv.putText(min500frame, f"DEBUG {fps:0.2f} FPS", (
                    12, 20), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.putText(min500frame, f"DEBUG FrameTime {frameTime*1000:0.2f}ms",
                           (12, 60), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.putText(min500frame, f"DEBUG video speed {videoSpeed:0.2f}%", (
                    12, 100), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
                cv.imshow('video EVEN MORE optimized', min500frame)


                totalTime = time.time() - startTime

                # print("REEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE\n")
                # print(resizeTime*100/totalTime)
                # print("REEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE\n")

                if cv.waitKey(1) & 0xFF == ord('q'):
                    bool_DisplayingVideo = False
                    continue

        print(f"DEBUG Video length {time.time() - debug_startTime:0.2f}")
        print(
            f"DEBUG Average fps {debug_frameCount / (time.time() - debug_startTime):0.2f} ")

    # def correctSpeed_displayEvenMoreOptimizedMethod(self):
    #     """
    #     SHOW THEM THE BEST OF YOU
    #     This method uses the facial detection pipeline (macroFacialDetection) written in the ImageManager class (look for eyes, find the pairs, crop out potential faces to run haarcascade facial detection on that area)
    #     By nature, this method is much more computationally heavier than just running the haarcascade facial detection directly on the image. So instead of running it on every frame, this method only
    #     run once its previous call has finished running. This will result in choppy face detection framerate (8 -> 15 fps).

    #     To improve the choppy framerate, I use past locations and run basic haarcascade facial detection (microFacialDetection) on those areas only. Then run eye detection in the new detected faces to update the angles of the faces.
    #     """
    #     # currentDetectedFacesManager is an array that store the tuples of most recent position of the detected faces and an int indicating to whether should the face be displayed on the frame (0 means yes, > 0 means no)
    #     currentDetectedFacesManager = []
    #     # while shadowDetectedFaces store the faces detected by macro facial detection, which lags 2-6 frames behind
    #     shadowDetectedFaces = []
    #     # outputQueue is used for storing threaded functions output
    #     outputQueue = queue.Queue()
    #     # this is a thread
    #     faceDetectionThread = threading.Thread()
    #     faceDetectionThread.start()

    #     # frameTimeInterval, prev_frame, and next_frame are used to determine the frame rate of the video, how often do we check for new frames
    #     frameTimeInterval = 1/self.videoFps
    #     prev_frame = time.time()
    #     next_frame = prev_frame + frameTimeInterval

    #     # actual_prev_frame and actual_next_frame are for determining the actual fps/frame time when displayed. The deviation from the actual fps/frametime values is because
    #     # of rendering and image processings applied to the image
    #     actual_prev_frame = prev_frame
    #     actual_next_frame = 0

    #     bool_DisplayingVideo = True

    #     # startTime, endTime, frameCount are used for debugging (checking average fps)
    #     debug_startTime = time.time()
    #     debug_frameCount = 0
    #     debug_frameSkip = 0

    #     while bool_DisplayingVideo:

    #         # print(f"DEBUG currentDetectedFacesManager length {len(currentDetectedFacesManager)}")
    #         # print(f"DEBUG shadowDetectedFaces length {len(shadowDetectedFaces)}")

    #         # Capture frame-by-frame
    #         ret, frame = self.video.read()

    #             # if frame is read correctly/ new frame is available, then ret is True.
    #         if not ret:
    #             print("Can't receive frame. Exiting ...")
    #             bool_DisplayingVideo = False
    #             continue

    #         # if its next time for the next frame to be drawn
    #         if time.time() >= next_frame:

    #         # resizing the frame to fit my tiny monitor, and to increase haarcascade detection performance too ye
    #         min500frame = resizeMinTo500(frame)

    #         # set new display time target (next_frame) after check if new frame is available instead of before. This is to make sure when live capturing
    #         # the newest image will be shown as soon as they are ready (variable fps maybe). If this was to be put before checking if new frame is available, newest
    #         # frame have to wait till the next frameTimeInterval to be shown if it misses the frame display time target, which increases latency (MAYBE :thonking:?)
    #         prev_frame = time.time()
    #         next_frame = prev_frame + frameTimeInterval

    #         # Our operations on the frame come here
    #         debug_frameCount = debug_frameCount + 1

    #         # variable face detection scanning fps
    #         if not faceDetectionThread.is_alive():
    #             # --- MACRO FACIAL DETECTION SECTION --- MACRO FACIAL DETECTION SECTION --- MACRO FACIAL DETECTION SECTION --- MACRO FACIAL DETECTION SECTION --- MACRO FACIAL DETECTION SECTION --- #

    #             # If the previous thoroughly facial detection (my implementation) has finished running
    #             # print(f"DEBUG frameSkip {debug_frameSkip + 1} frames")
    #             debug_frameSkip = 0

    #             # delayDetectedFaces is the array that store detectedArea objs returned by previously called findFacesCounterClockwiseMultipleAngles function.
    #             # These objs dont represent the current faces, but faces found in 2-6 frames previously when the function was called
    #             delayDetectedFaces = []
    #             # getting the most recent thoroughly detected faces' positions
    #             if not outputQueue.empty():
    #                 delayDetectedFaces = outputQueue.get()

    #                 # print(f"DEBUG delayDetectedFaces length {len(delayDetectedFaces)}")

    #             # Merging the macro facial detection into the micro facial detection list
    #             if len(shadowDetectedFaces) == 0:
    #                 shadowDetectedFaces = delayDetectedFaces
    #                 for shadowFace in shadowDetectedFaces:
    #                     currentDetectedFacesManager.append([shadowFace.copy(), 0])
    #             elif len(delayDetectedFaces) != 0:

    #                 # check if there's a new face detected by macro facial detection
    #                 for shadowFace in shadowDetectedFaces:

    #                     bool_innerLoop = False
    #                     i = 0
    #                     if i < len(delayDetectedFaces):
    #                         bool_innerLoop = True

    #                     while bool_innerLoop:
    #                         delayFace = delayDetectedFaces[i]
    #                         # if the face returned by macro facial detection has already exists then pop it off
    #                         if shadowFace.similarSize(delayFace, self.HARDCODED_similarSizeScale) and shadowFace.overlap(delayFace):
    #                             delayDetectedFaces.pop(i)
    #                             i = i - 1
    #                         i = i + 1
    #                         if i >= len(delayDetectedFaces):
    #                             bool_innerLoop = False

    #                 # if there are faces that haven't been detected then add them to the currentDetectedFacesManager list
    #                 if len(delayDetectedFaces) != 0:
    #                     for newFace in delayDetectedFaces:
    #                         currentDetectedFacesManager.append([newFace, 0])

    #             # Run the next macro facial detection (my implementation) in a seperate thread to avoid interfering with updating new frames
    #             frameMgnr = ImageManager(min500frame)
    #             args = [(45, 0, -45), 1.1, 10]
    #             faceDetectionThread = threading.Thread(target = self.HELPER_findFacesCounterClockwiseMultipleAngles_to_outputQueue, args= [frameMgnr, args, outputQueue])
    #             faceDetectionThread.start()

    #             # update the shadowDetectedFace list for when the thread finished running
    #             for currentFaceManager in currentDetectedFacesManager:
    #                 currentFace = currentFaceManager[0]
    #                 shadowDetectedFaces.append(currentFace.copy())

    #         else:

    #             debug_frameSkip = debug_frameSkip + 1

    #             # --- MICRO FACIAL DETECTION SECTION --- MICRO FACIAL DETECTION SECTION --- MICRO FACIAL DETECTION SECTION --- MICRO FACIAL DETECTION SECTION --- MICRO FACIAL DETECTION SECTION --- #

    #             # updating the detected faces in the last frame by searching around their area
    #             if len(currentDetectedFacesManager) != 0:
    #                 bool_loop = True
    #                 index_currentDetectedFacesManager = 0
    #                 if index_currentDetectedFacesManager >= len(currentDetectedFacesManager):
    #                     bool_loop = False
    #                 while bool_loop:
    #                     face = currentDetectedFacesManager[index_currentDetectedFacesManager][0]

    #                     dimensions = face.dimensions
    #                     counterClockwiseAngle = face.counterClockwiseAngle
    #                     faceCenter = Point(face.center.x, face.center.y)

    #                     # rotate clockwise because the face's angle is the counter clockwise angle of the right eye to the left eye
    #                     frameOrigin = Point(min500frame.shape[1]/2, min500frame.shape[0]/2)
    #                     rotatedFrame = rotateClockwise(min500frame, counterClockwiseAngle)
    #                     rotatedFrameOrigin = Point(rotatedFrame.shape[1]/2, rotatedFrame.shape[0]/2)

    #                     # rotate clockwise the faceCenter and project it to the new image
    #                     rotatedFaceCenter = faceCenter.rotatePointClockwise(frameOrigin, counterClockwiseAngle)
    #                     rotatedFaceCenter = rotatedFaceCenter.projectPoint(frameOrigin, rotatedFrameOrigin)

    #                     # Calculate the region of the potential face
    #                     searchingDimensions = (dimensions[0] * self.HARDCODED_updateFaceLocationSearchMultiplier, dimensions[1] * self.HARDCODED_updateFaceLocationSearchMultiplier)
    #                     cropRangeMinY = int(max(0, rotatedFaceCenter.y - searchingDimensions[1] / 2))
    #                     cropRangeMaxY = int(min(rotatedFrame.shape[0], rotatedFaceCenter.y + searchingDimensions[1] / 2))

    #                     cropRangeMinX = int(max(0, rotatedFaceCenter.x - searchingDimensions[0] / 2))
    #                     cropRangeMaxX = int(min(rotatedFrame.shape[1], rotatedFaceCenter.x + searchingDimensions[0] / 2 ))

    #                     # Cropping out the region of the potential face
    #                     try:
    #                         croppedPotentialFace = np.zeros((cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX), dtype='uint8')
    #                         croppedPotentialFace[0: croppedPotentialFace.shape[0], 0: croppedPotentialFace.shape[1]] = \
    #                             rotatedFrame[cropRangeMinY:cropRangeMaxY, cropRangeMinX:cropRangeMaxX]

    #                     except:
    #                         croppedPotentialFace = np.zeros((cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX, 3), dtype ='uint8')
    #                         croppedPotentialFace[0: croppedPotentialFace.shape[0], 0: croppedPotentialFace.shape[1]] = \
    #                             rotatedFrame[cropRangeMinY:cropRangeMaxY, cropRangeMinX:cropRangeMaxX]

    #                     croppedPotentialFaceCenter = Point(croppedPotentialFace.shape[1]/2 , croppedPotentialFace.shape[0]/2)

    #                     # gray scale the image
    #                     grayCroppedPotentialFace = cv.cvtColor(croppedPotentialFace, cv.COLOR_BGR2GRAY)
    #                     # run basic haarcascade facial detection on it
    #                     # CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ? CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ? CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ?
    #                     detectedFaces = haarcascade_FacialDetection(grayCroppedPotentialFace, 1.3, 6)

    #                     # if found nothing then move to the next face's last frame location
    #                     if len(detectedFaces) == 0:

    #                         # increment the faceNotFoundCount
    #                         faceNotFoundCount = currentDetectedFacesManager[index_currentDetectedFacesManager][1]
    #                         faceNotFoundCount = faceNotFoundCount + 1
    #                         if faceNotFoundCount >= self.HARDCODED_faceNotFoundCountLimit:

    #                             currentDetectedFacesManager.pop(index_currentDetectedFacesManager)
    #                             # if there's still more currentFaceManager to iterate
    #                             if index_currentDetectedFacesManager < len(currentDetectedFacesManager):
    #                                 continue

    #                             # if there's no more currentFaceManager to iterate then break the loop
    #                             bool_loop = False
    #                             continue

    #                         # update faceNotFoundCount
    #                         currentDetectedFacesManager[index_currentDetectedFacesManager][1] = faceNotFoundCount

    #                         # if still under the faceNotFoundLimit then continue iterating
    #                         continue

    #                     # if found anything, then convert the rectangle coordinates received from detectMultiScale function to a detectedFace obj.
    #                     biggestFace = detectedFaces[0]
    #                     for i in range(1, len(detectedFaces)):
    #                         # comparing the dimensions (width only, cause if width is > then height is also >) among all the faces found.
    #                         if biggestFace[2] < detectedFaces[i][2]:
    #                             biggestFace = detectedFaces[i]

    #                     # Creating DetectedFace obj representing the biggest face found in the potential region above
    #                     biggestFace = DetectedFace((biggestFace[0], biggestFace[1]), (biggestFace[2], biggestFace[3]))

    #                     # Finding biggestFace's angle
    #                     # CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ? CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ? CAN IMPROVE PERFORMANCE BY ADDING APPROPRIATE MINSIZE AND MAXSIZE ?

    #                     # find the eyes
    #                     tempEyesArray = haarcascade_EyeDetection(grayCroppedPotentialFace, 1.3, 10)
    #                     detectedEyes = []

    #                     if face.leftEye != None:
    #                         referenceEye = face.leftEye
    #                     elif face.rightEye != None:
    #                         referenceEye = face.rightEye
    #                     else:
    #                         faceUpperLeftPoint = biggestFace.upperLeft
    #                         # if the face has no eyes found, then create a standard reference eye
    #                         # HARDCODED ___ Eye's width is usually at most 3/8 of face's width. Eye's height is usually at most 1/3 of face's height
    #                         referenceEye = DetectedArea((faceUpperLeftPoint.x + biggestFace.dimensions[0]/8, faceUpperLeftPoint.y + biggestFace.dimensions[1]/6), (biggestFace.dimensions[0]* 3/ 8, biggestFace.dimensions[1]/3) )

    #                     for (x, y, w, h)in tempEyesArray:
    #                         eye = DetectedArea((x, y), (w, h))
    #                         # If the shape and size of the eye found in the image is approximately the same as the reference eye
    #                         if eye.similarSize(referenceEye, self.HARDCODED_similarSizeScale):
    #                             detectedEyes.append(eye)

    #                     # find the pairs of eyes
    #                     if len(detectedEyes) > 1:
    #                         # This might be slightlyyyyy overkill but usually there's like at most 2-3 eyes so it wont be too computationally heavy.
    #                         for i in range(len(detectedEyes)):
    #                             j = i + 1
    #                             for j in range(len(detectedEyes)):
    #                                 if detectedEyes[i].appropriateDistanceTo(detectedEyes[j], self.HARDCODED_pairOfEyesDistanceRange[0], self.HARDCODED_pairOfEyesDistanceRange[1]):
    #                                     leftEye = detectedEyes[i]
    #                                     rightEye = detectedEyes[j]
    #                                     if detectedEyes[i].center.x > detectedEyes[j].center.x:
    #                                         leftEye = detectedEyes[j]
    #                                         rightEye = detectedEyes[i]
    #                                     # if the distance of LeftEye's center to face's upperLeft Point ~ that of RightEye's center to face's upperRight Point
    #                                     # and distance of LeftEye's center to face's lowerLeft Point ~ that of RightEye's center to face's lowerRight Point
    #                                     # then this is probably the right pair of eyes
    #                                     upperLeftDist = leftEye.center.distTo(biggestFace.upperLeft)
    #                                     upperRightDist = rightEye.center.distTo(biggestFace.upperRight)
    #                                     lowerLeftDist = leftEye.center.distTo(biggestFace.lowerLeft)
    #                                     lowerRightDist = rightEye.center.distTo(biggestFace.lowerRight)
    #                                     criterias = 0
    #                                     if upperLeftDist < upperRightDist and upperLeftDist > upperRightDist * self.HARDCODED_distanceVariation_EyeToCorners_MinOfMax:
    #                                         criterias = criterias + 1
    #                                     if upperRightDist <= upperLeftDist and upperRightDist > upperLeftDist * self.HARDCODED_distanceVariation_EyeToCorners_MinOfMax:
    #                                         criterias = criterias + 1
    #                                     if lowerLeftDist < lowerRightDist and lowerLeftDist > lowerRightDist * self.HARDCODED_distanceVariation_EyeToCorners_MinOfMax:
    #                                         criterias = criterias + 1
    #                                     if lowerRightDist <= lowerLeftDist and lowerRightDist > lowerLeftDist * self.HARDCODED_distanceVariation_EyeToCorners_MinOfMax:
    #                                         criterias = criterias + 1
    #                                     if criterias == 2:
    #                                         biggestFace.leftEye = leftEye
    #                                         biggestFace.rightEye = rightEye
    #                                         break

    #                     # find the face's angle using the pair of eyes
    #                     if biggestFace.leftEye != None and biggestFace.rightEye != None:
    #                         # calculate teh relative angle created by the two eyes
    #                         relativeCounterClockwiseAngle = biggestFace.leftEye.center.relativeCounterClockwiseAngle(biggestFace.rightEye.center)
    #                         # the face's current angle is the already rotated angle pre-haarcascade detection + the relative angle
    #                         biggestFace.counterClockwiseAngle = counterClockwiseAngle + relativeCounterClockwiseAngle
    #                         biggestFace.rotateAreaCounterClockwise(biggestFace.center, relativeCounterClockwiseAngle)
    #                     else:
    #                         # if there's not a pair of eyes then the face's angle = the previous face that was detected in the same area's angle
    #                         biggestFace.counterClockwiseAngle = counterClockwiseAngle

    #                     # Returning the face's coordinates to the original frame
    #                     biggestFace.projectArea(croppedPotentialFaceCenter, rotatedFaceCenter)
    #                     biggestFace.rotateAreaCounterClockwise(rotatedFrameOrigin, biggestFace.counterClockwiseAngle)
    #                     biggestFace.projectArea(rotatedFrameOrigin, frameOrigin)

    #                     # update the face's location and reset faceNotFoundCount value to 0
    #                     currentDetectedFacesManager[index_currentDetectedFacesManager][0] = biggestFace
    #                     currentDetectedFacesManager[index_currentDetectedFacesManager][1] = 0

    #                     # continue iterating the currentFaceManager
    #                     index_currentDetectedFacesManager = index_currentDetectedFacesManager + 1
    #                     if index_currentDetectedFacesManager >= len(currentDetectedFacesManager):
    #                         bool_loop = False

    #         for faceManager in currentDetectedFacesManager:
    #             face = faceManager[0]
    #             face.draw(min500frame, (255, 255, 255), 2)

    #         # Calculating fps, timeFrame, video speed
    #         actual_next_frame = time.time()
    #         frameTime = actual_next_frame - actual_prev_frame

    #         actual_prev_frame = actual_next_frame
    #         fps = 1/frameTime
    #         videoSpeed = frameTimeInterval * 100 / frameTime

    #         # Display the resulting frame
    #         cv.putText(min500frame, f"DEBUG {fps:0.2f} FPS", (12, 20), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
    #         cv.putText(min500frame, f"DEBUG FrameTime {frameTime*1000:0.2f}ms", (12, 60), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
    #         cv.putText(min500frame, f"DEBUG video speed {videoSpeed:0.2f}%", (12, 100), cv.FONT_HERSHEY_SIMPLEX, 0.6, (255, 255, 0), 2)
    #         cv.imshow('video EVEN MORE optimized', min500frame)

    #         if cv.waitKey(1) & 0xFF == ord('q'):
    #             bool_DisplayingVideo = False
    #             continue

    #     print(f"DEBUG Video length {time.time() - debug_startTime:0.2f}")
    #     print(f"DEBUG Average fps {debug_frameCount / (time.time() - debug_startTime):0.2f} ")

    def HELPER_findFacesCounterClockwiseMultipleAngles_to_outputQueue(self, frameManager, arguments, outputQueue):
        """
        Help multithreading a function by being a dummy function which only purpose is to call the threadedly-desired? function and save its output into the
        argument queue instead of returning it.
            :param function: the function that needs to be multithreaded
            :param arguments: a list of function's arguments []
            :param outputQueue: the queue that stores the function output
        """
        outputQueue.put(frameManager.findFacesCounterClockwiseMultipleAngles(
            arguments[0], arguments[1], arguments[2]))



  


import cv2 as cv
from numpy import append
from . import DetectedArea
# from .DetectedArea import DetectedFace
from . import ImageManager
from . import Point
from .HelperFunctions import *
import os





def findAndDrawFacesVanilla(frame):
    """
    Create a grayscaled version of the image, find the faces, and draw them on the original image
    All using vanilla facial detection that came with opencv
    """
    # package_directory = os.path.dirname(__file__)
    # path = os.path.join(package_directory, "classifier\haarcascade_frontalface_default.xml")
    path = os.path.join(os.getcwd(), "classifier/hasarcascade_frontalface_default.xml")
    print(path +" ------------------------------------------------------------")
    haar_cascasde_face = cv.CascadeClassifier(path)



    grayFrame = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)

    faces = haar_cascasde_face.detectMultiScale(
        grayFrame, 1.1, 10, minSize=(120, 120))

    for (x, y, w, h) in faces:
        cv.rectangle(frame, (x, y), (x+w, y+h), (255, 255, 0), 2)
        

      

  


  


from threading import Thread
import threading

class FrameThread(Thread):
    def __init__(self, threadID, frameID, customFunction, frame, frameLock, endPoint):
        threading.Thread.__init__(self)
        self.threadID = threadID
        self.frameID = frameID
        self.task = customFunction
        self.frame = frame
        self.frameLock = frameLock
        self.endPoint = endPoint

    def run(self):
        self.frame = self.task(self.frame)
        # Stop thread from stop running => changing the VideoManager frameThread array status
        # Affecting the EmittingThread from getting correct values and send to the clients
        # self.frameLock.acquire()
        # self.frameLock.release()
        




from threading import Thread, Lock
from .FrameThread import FrameThread
import time

class ServerVideoManager:
    def __init__(self, frameRate, socket, clientID, grayScale):
        self.testingGrayScale = grayScale

        # Used for the emitting thread
        self.playing = True
        self.frameRate = frameRate

        self.frameThreads = [None, None, None, None, None, None, None, None, None, None]
        self.freeThreadIDs = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        self.threadCount = 10
        self.returnedFrameID = 0
        
        self.socket = socket
        self.clientID = clientID

        self.emittingThread = None
        self.frameLock = Lock()


    def start(self):
        self.playing = True
        self.emittingThread = Thread(target=emitter, args=[self])
        self.emittingThread.start()
        
    def stop(self):
        self.playing = False



    def processNextFrame(self, customFunction, frame, frameID, endPoint):
        # Its pretty impossible to have all 5 threads taken up
        if not len(self.freeThreadIDs):
            print('*****************************************************************************************')
            print('*  ERROR: in VideoManager file: No more free threads. Sir can I have sum more threads?  *')
            print('*****************************************************************************************')
            return

        threadID = self.freeThreadIDs.pop(0)

        self.frameThreads[threadID] = FrameThread(threadID, frameID, customFunction, frame, self.frameLock, endPoint)
        self.frameThreads[threadID].start()
        
    
    



def emitter(videoManager):
    
    while videoManager.playing: 
        if len(videoManager.freeThreadIDs) < videoManager.threadCount:
            

            i = 0
            frameThreads = videoManager.frameThreads
            freeThreadIDs = videoManager.freeThreadIDs
            finishedThreadIDs = []


            videoManager.frameLock.acquire()
            
            while i < len(frameThreads):

                # If thread t has finished running
                
                if frameThreads[i] != None and not frameThreads[i].is_alive():
                    # add the finished thread id onto the freeThreadIDs attribute of the videoManager
                    freeThreadIDs.append(i)

                    # add the finished thread id onto the finishedThreadArray thats gonna be used later
                    finishedThreadIDs.append(i)

                i += 1            

            videoManager.frameLock.release()
            

            if len(finishedThreadIDs) >= 1:
                # If more than one thread finished, then only send the latest one's output to the client
                
                latestFrameIDIndex = 0
                latestFrameID = frameThreads[finishedThreadIDs[latestFrameIDIndex]].frameID
                

                # Finding the newest frame that just got rendered by frameThread
                i = 1
                while i < len(finishedThreadIDs):

                    if(frameThreads[finishedThreadIDs[i]].frameID > latestFrameID):
                        latestFrameIDIndex = i
                        latestFrameID = frameThreads[finishedThreadIDs[i]].frameID

                    i += 1

                # if the latest frame that just got rendered by frameThread is newer than the last frame that we sent back to the client
                # then send that latest frame to the client
                if latestFrameID > videoManager.returnedFrameID: 
                    temp = frameThreads[finishedThreadIDs[latestFrameIDIndex]]
                    videoManager.returnedFrameID = latestFrameID
                    data = {'base64_responseFrame': temp.frame, 'frameID': temp.frameID}
                    videoManager.socket.emit(temp.endPoint, data, to=videoManager.clientID)

                # if the latest frame is behind the last frame that was sent to the client, then do nothing
                


                # Removing the finished threads out of the running threads array
                for i in finishedThreadIDs:
                    frameThreads[i] = None

            time.sleep(1/(3*videoManager.frameRate))
        else:
            time.sleep(1/(6 * videoManager.frameRate))
                






import math
import cv2 as cv
import numpy as np

haar_cascasde_eye = cv.CascadeClassifier("classifier/haarcascade_eye.xml")
haar_cascasde_nose = cv.CascadeClassifier("classifier/haarcascade_nose.xml")
# Default frontal face xml => more accurate but slower
# haar_cascasde_face = cv.CascadeClassifier("classifier/haarcascade_frontalface_default.xml")
# Lighter frontal face xml => less accurate (accurate enuf) but faster
haar_cascasde_face = cv.CascadeClassifier("classifier/lbpcascaade_frontalface_improved.xml")


# Haarcascade basic detection
def haarcascade_FacialDetection(image, scaleFactor, minNeighbors, minSize = None, maxSize = None):
    """
    This is to be used by videoManager only because there seems to be some error when used in imageManager (caused by multithreading probably)
    Just run the haarcascade facial detection on the image without modifying the input image
        :return the detectMultiScale output
    """
    if minSize == None:
        if maxSize == None:
            return haar_cascasde_face.detectMultiScale(image, scaleFactor, minNeighbors)
        return haar_cascasde_face.detectMultiScale(image, scaleFactor, minNeighbors, maxSize = maxSize)
    if maxSize == None:
        return haar_cascasde_face.detectMultiScale(image, scaleFactor, minNeighbors, minSize = minSize)
    return haar_cascasde_face.detectMultiScale(image, scaleFactor, minNeighbors, minSize = minSize, maxSize = maxSize)


def haarcascade_EyeDetection(image, scaleFactor, minNeighbors, minSize = None, maxSize = None):
    """
    Just run the haarcascade eye detection on the image without modifying the input image
        :return the detectMultiScale output
    """
    if minSize == None:
        if maxSize == None:
            return haar_cascasde_eye.detectMultiScale(image, scaleFactor, minNeighbors)
        return haar_cascasde_eye.detectMultiScale(image, scaleFactor, minNeighbors, maxSize = maxSize)
    if maxSize == None:
        return haar_cascasde_eye.detectMultiScale(image, scaleFactor, minNeighbors, minSize = minSize)
    return haar_cascasde_eye.detectMultiScale(image, scaleFactor, minNeighbors, minSize = minSize, maxSize = maxSize)
    


# Rotation
def rotateClockwise(img, angle):
    """
    This functions return a clockwise rotated image without cropping any part of the original image and keeping the center of the original 
    image at the center of the new rotated image. The returned image may have larger dimensions than the original but as small as possible.
        :param img: the original image
        :param angle: the rotating angle clockwise
        :return: the rotated-around-the-center image without cropping
    """
    angle = -angle
    return rotateCounterClockwise(img, angle)


def rotateCounterClockwise(img, angle):
    """
    This functions return a counter-clockwise rotated image without cropping any part of the original image and keeping the center of the original 
    image at the center of the new rotated image. The returned image may have larger dimensions than the original but as small as possible.
        :param img: the original image
        :param angle: the rotating angle counter-clockwise
        :return: the rotated-around-the-center image without cropping
    """
    (height, width) = img.shape[:2]

    # Calculating the dimension of the new canvas to store the entire rotated image without cropping, diagAngleA is used to calculate fittingWidth, diagAngleB is for fittingHeight
    diagLen = math.sqrt(width**2 + height**2)
    diagAngleA = math.atan(width/height)
    diagAngleB = math.pi/2 - diagAngleA

    # TempAngle is used to keep the angle value in math.cos stays between -90 and 90 degree => consistence cos value 
    # cos value is the cosine of the offset of frame/img diagonal line from the x/y axis (diagAngle + tempAngle*math.pi/180 - math.pi/2). 
    # And using the value of angle, we will know which cos value is used to determined width and height.
    # angle < 360 to avoid edge cases
    angle = angle % 360
    tempAngle = angle%90
    if angle == 0 or angle == 180:
        fittingWidth = width
        fittingHeight = height
    elif angle == 90 or angle == 270:
        fittingWidth = height
        fittingHeight = width
    elif (angle > 0 and angle < 90) or (angle > 180 and angle < 270):
        
        fittingWidth = math.floor(math.cos(diagAngleA + tempAngle*math.pi/180 - math.pi/2) * diagLen)
        fittingHeight = math.floor(math.cos(diagAngleB + tempAngle*math.pi/180 - math.pi/2) * diagLen)
    else :
        fittingWidth = math.floor(math.cos(diagAngleB + tempAngle*math.pi/180 - math.pi/2) * diagLen)
        fittingHeight = math.floor(math.cos(diagAngleA + tempAngle*math.pi/180 - math.pi/2) * diagLen)

    # Drawing pre rotated image ontop of bigger canvas
    # rotatedCanvas initially takes the largest horizontal/vertical value among newWidth, newHeight, img.shape[1], img.shape[2] to prevent the image from cropping and index out of bound
    # Once the image is rotated on rotatedCanvas, we will crop out the excess part
    try:
        rotatedCanvas = np.zeros((max(fittingHeight, img.shape[0]), max(fittingWidth, img.shape[1]), 3), dtype='uint8')
        rotatedCanvas[math.floor(rotatedCanvas.shape[0]/2 - img.shape[0]/2): math.floor(rotatedCanvas.shape[0]/2 + img.shape[0]/2), math.floor(rotatedCanvas.shape[1]/2 - img.shape[1]/2): math.floor(rotatedCanvas.shape[1]/2 + img.shape[1]/2)] = img
    except:
        rotatedCanvas = np.zeros((max(fittingHeight, img.shape[0]), max(fittingWidth, img.shape[1])), dtype='uint8')
        rotatedCanvas[math.floor(rotatedCanvas.shape[0]/2 - img.shape[0]/2): math.floor(rotatedCanvas.shape[0]/2 + img.shape[0]/2), math.floor(rotatedCanvas.shape[1]/2 - img.shape[1]/2): math.floor(rotatedCanvas.shape[1]/2 + img.shape[1]/2)] = img
    # rotatedCanvas[100:200,400:500] = 0,255,0
    # [upper height bound: lower height bound, left width bound: right width bound]
    # cv.imshow("Test draw on cavnas before rotate", rescaleFrame(rotatedCanvas, 500))

    #rotPoint (width, height)
    rotPoint = (rotatedCanvas.shape[1]//2, rotatedCanvas.shape[0]//2)
    
    rotMat = cv.getRotationMatrix2D(rotPoint, angle, 1.0)
    excessDimensions = (rotatedCanvas.shape[1], rotatedCanvas.shape[0])
    rotatedCanvas = cv.warpAffine(rotatedCanvas, rotMat, excessDimensions)
    # rotatedCanvas = rotatedCanvas[100:1200, 100:600]
    # Cropping excess borders
    rotatedCanvas = rotatedCanvas[math.floor((rotatedCanvas.shape[0] - fittingHeight)/2) : math.floor((rotatedCanvas.shape[0] + fittingHeight)/2), math.floor((rotatedCanvas.shape[1] - fittingWidth)/2) : math.floor((rotatedCanvas.shape[1] + fittingWidth)/2)]
    return rotatedCanvas





# Resize
def resizeMinTo500(img):
    """
    Create a resized clone of the original image such that the smaller dimension of the image = 500px, and the other dimension is kept to scale
        :param img: the image need to be resized
        :return a resized copy of the iamge
    """
    
    (height, width) = img.shape[:2]

    resizedImage = img.copy()

    if height < width:
        width = int(width * 500/ height)
        height = 500
        
    else:
        height = int(height * 500/ width)
        width = 500
        
    dimensions = (width, height)
    return cv.resize(resizedImage, dimensions, interpolation= cv.INTER_LINEAR)

# Resize
def resizeMinTo(img, size):
    """
    Create a resized clone of the original image such that the smaller dimension of the image = 500px, and the other dimension is kept to scale
        :param img: the image need to be resized
        :return a resized copy of the iamge
    """
    size = int(size)
    (height, width) = img.shape[:2]

    resizedImage = img.copy()

    if height < width:
        width = int(width * size/ height)
        height = size
        
    else:
        height = int(height * size/ width)
        width = size
        
    dimensions = (width, height)
    return cv.resize(resizedImage, dimensions, interpolation=cv.INTER_LINEAR)


# Resize TEST INTERPOLATION SPEED
def TEST_INTERPOLATION_TYPE_resizeMinTo(img, size, interpolationType):
    """
    Create a resized clone of the original image such that the smaller dimension of the image = 500px, and the other dimension is kept to scale
        :param img: the image need to be resized
        :param interapolationType: the interpolation type with
            0: INTER_AREA
            1: INTER_CUBIC
            2: INTER_LANCZOS4
            3: INTER_NEAREST
            4: INTER_LINEAR
        :return a resized copy of the iamge
    """
    size = int(size)
    (height, width) = img.shape[:2]

    resizedImage = img.copy()

    if height < width:
        width = int(width * size/ height)
        height = size
        
    else:
        height = int(height * size/ width)
        width = size
        
    dimensions = (width, height)
    if interpolationType == 0:
        return cv.resize(resizedImage, dimensions, interpolation=cv.INTER_AREA)
    elif interpolationType == 1:
        return cv.resize(resizedImage, dimensions, interpolation=cv.INTER_CUBIC)
    elif interpolationType == 2:
        return cv.resize(resizedImage, dimensions, interpolation=cv.INTER_LANCZOS4)
    elif interpolationType == 3:
        return cv.resize(resizedImage, dimensions, interpolation=cv.INTER_NEAREST)
    elif interpolationType == 4:
        return cv.resize(resizedImage, dimensions, interpolation=cv.INTER_LINEAR)




import math


class Point:
    """ 
    Point stores information of a point in an image and can calculate distance from itself to other Points.
    """
    def __init__(self, xCoord, yCoord, angle = 0):
        """ 
        Construct a point object.
            :param xCoord: the x coordinate.
            :param yCoord: the y coordinate.
        """
        self.x = xCoord
        self.y = yCoord

    def copy(self):
        """
        Return a deep copy of itself
            :return a deep copy of itself
        """
        return Point(self.x, self.y)
        
    def exportCoordinates(self):
        """
        Return a tuple containing the x and y coordinates of the Point object
            :return the tuple containing x and y coordinates
        """
        return (int(self.getX()), int(self.getY()))

    def getX(self):
        """
        Return the point's x coordinate
            :return x coordinate
        """
        return self.x
        
    def getY(self):
        """
        Return the point's y coordinate
            :return y coordinate
        """
        return self.y

    def distTo(self, otherPoint):
        """
        Calculate the distance from the current point to the other point.
            :return the distance between two points.
        """
        distance = math.sqrt((self.x - otherPoint.x)**2 + (self.y - otherPoint.y)**2)
        return distance

    def rotatePointCounterClockwise(self, origin, angle):
        """
        Rotate the current point object by angle value counter-clockwise relative to the given origin.
            :param origin: the point where the function caller point object is going to rotate around
            :param angle: the angle the function caller point object is going to be rotated by
            :return the rotated point as a separated object
        """

        angle = angle % 360
        if angle == 0: 
            return Point(self.x, self.y)

        # Calculating the relative angle to the origin
        dist = self.distTo(origin)
        deltaX = self.x - origin.x
        deltaY = self.y - origin.y
        
        
        if deltaX > 0: 
            # if the point is to the right of the origin
            relativeAngle = (360 - (math.atan(deltaY/deltaX) * 180/math.pi)) % 360
        elif deltaX < 0:
            # if the point is to the left of the origin
            relativeAngle = 180 - (math.atan(deltaY/deltaX) * 180/math.pi)
        elif deltaX == 0:
            if deltaY > 0:
                relativeAngle = 270
            elif deltaY < 0:
                relativeAngle = 90
            else:
                # the point is the origin, no need to translate
                return Point(self.x, self.y)
        

        relativeAngle = (relativeAngle + angle + 360) % 360
        deltaX = math.cos(relativeAngle * math.pi / 180) * dist
        # negative sign cause sin positive when deltaY negative and vice versa
        deltaY = - math.sin(relativeAngle * math.pi / 180) * dist
         

        return Point(origin.x + deltaX, origin.y + deltaY)
        
    
    def rotatePointClockwise(self, origin, angle):
        """
        Translate the current point object by angle value clockwise relative to the given origin.
            :param origin: the point where the function caller point object is going to rotate around
            :param angle: the angle the function caller point object is going to be rotated by
            :return the rotatedPoint as a separated object
        """
        return self.rotatePointCounterClockwise(origin, -angle)

    def projectPoint(self, oldOrigin, newOrigin):
        """
        Calculate the position of a new point such that its position relative to newOrigin is the same as its 
        current relative position to old Origin.
        exp: current point is (1,1), oldOrigin is (0,0), and new origin is (2,0) => returned point is (3,1)
            :param oldOrigin: the old origin point
            :param newOrigin: the projected old Origin
            :return the projected current point such that its position relative to newOrigin is the same as current point to oldOrigin
        """
        projectedX = self.x + newOrigin.x - oldOrigin.x
        projectedY = self.y + newOrigin.y - oldOrigin.y
        return Point(projectedX, projectedY)

    
    def relativeCounterClockwiseAngle(self, otherPoint):
        """
        Calculate the angle the otherPoint creates if the self Point was the origin.
        The function ONLY return positive angle. If it returns -1 that means otherPoint
        is the same as self.
            :param otherPoint: the other Point object
            :return the angle in degree
        """
        deltaX = otherPoint.x - self.x
        deltaY = otherPoint.y - self.y
        
        if deltaX > 0: 
            # if the point is to the right of the origin
            relativeAngle = (360 - (math.atan(deltaY/deltaX) * 180/math.pi)) % 360
        elif deltaX < 0:
            # if the point is to the left of the origin
            relativeAngle = 180 - (math.atan(deltaY/deltaX) * 180/math.pi)
        elif deltaX == 0:
            if deltaY > 0:
                relativeAngle = 270
            elif deltaY < 0:
                relativeAngle = 90
            else:
                # the point is the origin, no need to translate
                return -1
        return relativeAngle



import cv2 as cv
from . import Point
from . import DetectedArea
from .DetectedArea import DetectedFace
from .HelperFunctions import *
import numpy as np


""" 
    At this point in development, what I need the ImageManager to do is store the image, perform eyes detection on the image in 0, 90, 180, and 270 degree.
    
    Then ImageManager collects all "detected eyes" called DetectedArea objs, which still has the x, y cordinate corresponding to the rotated angle of the image when
    they were found. It then proceed to convert the x, y value to the cordinate in the original non-rotated image.
    
    After that, it checks for overlapping DetectedAreas and merge them into one big DetectedArea obj.

    Once that's done, we get an array of all the location of the eyes and their sizes. We can then connect every possible pair of approximately-same-size-eyes to see 
    if the distance between them are "reasonable". If a pair of eyes has a reasonable distance inbetweeen and the size of the eyes don't differ too much, 
    we can call it a potential face.

    Once we get an array of potential faces, we rotate the image by an angle decided by the position of the two eyes (to increase precision), crop out the face (to increase performance), 
    and run facial detection on that.

    At the end, ImageManger should have an array list of all detected faces.

"""


class ImageManager:
    """
    A manager that stores the actual image and can do image processing function on it. This object will be used to take care of facial detection.
    """
    haarcascade_eye = cv.CascadeClassifier("classifier/haarcascade_eye.xml")
    haarcascade_face = cv.CascadeClassifier("classifier/haarcascade_frontalface_default.xml")
    # haarcascade_face = cv.CascadeClassifier("classifier/lbpcascaade_frontalface_improved.xml")
    haarcascade_nose = cv.CascadeClassifier("classifier/haarcascade_nose.xml")
    HARDCODED_similarSizeScale = 0.7
    HARDCODED_pairOfEyesDistanceRange = (1.5, 3.5)
    # eye and face min and max dimensions in a 500pixel x ? pixel images
    HARDCODED_eyeMinDimensions = (10, 10)
    HARDCODED_eyeMaxDimensions = (200, 200)
    HARDCODED_faceMinDimensions = (80, 80)
    HARDCODED_faceMaxDimensions = (500, 500)

    def __init__(self, img):
        """
        Constructing an ImageManger object
            :param img: the image/frame that we will run facial detection on
        """
        # Blank canvas that we are going to use to store the rotated image
        self.image = img
        self.grayImage = cv.cvtColor(self.image, cv.COLOR_BGR2GRAY)
        self.imageCenter = Point(img.shape[1]/2, img.shape[0]/2)
    
    

    
    def HELPER_runHaarDetectionCounterClockwiseAngle(self, detector, minDimensions, maxDimensions, angle, scaleFactor, minNeighbors):
        """
        Run the given haarDetection on the image rotated by the given angle and generate 1 array that 
        contain detected objects found.         
            :param detector: the haarcascade object that is going to scan the image
            :param angle: the angle by which the image is rotated
            :param scaleFactor: scaleFactor parameter for detectMultiScale function
            :param minNeighbors: minNeighbors parameter for detectMultiScale function
            :return a 2-tuple with the first element being an array containing all the raw coordinates of the detected objects in the rotated image,
            and the second element the Point object containing the coordinates of the center of the rotated image.
        """

        angle = angle % 360

        # finding objects in then given angle degree rotated image
       
        rotatedGrayImage = rotateCounterClockwise(self.grayImage, angle)
            
        # Collecting raw detected objects received from detectMultiScale
        rawObjs = detector.detectMultiScale(rotatedGrayImage, scaleFactor = scaleFactor, minNeighbors = minNeighbors, minSize = minDimensions, maxSize = maxDimensions)
            
        detectedAreas = []
        rotatedCenter = Point(rotatedGrayImage.shape[1]/2, rotatedGrayImage.shape[0]/2)

        for obj in rawObjs:
            # Convert raw information into DetectedArea obj
            area = DetectedArea((obj[0], obj[1]), (obj[2], obj[3]))

            # Put the translated DetectedArea into translatedObj array
            detectedAreas.append(area)

        return (detectedAreas, rotatedCenter)    


    def HELPER_rotateDetectedAreaClockwise(self, rawPositions, origin, angle):
        """
        Rotate detectedAreas in rawPositions around the given origin
            :param rawPositions: the array that contains detectedAreas with raw values taken from detectMultiScale
            :param origin: the origin which the detectedAreas are rotating around
            :param angle: the angle by which the image was rotated when detectMultiScale ran
            :return an array containing detectedAreas with translated coordinates in the non-rotated image
        """
        angle = angle % 360
        if angle == 0:
            return rawPositions

        # Translate the coordinates to the coordinates in the non-rotated image
        for area in rawPositions:
            area.rotateAreaClockwise(origin, angle)

        return rawPositions


    def HELPER_projectDetectedArea(self, rawPositions, rotatedCenter):
        """
        Project the raw positions such that the new positions relative to self.imageCenter is the same
        as relative positions of the old Coordinates to rotatedCenter.
            :param rawPositions: the array that contains detectedAreas
            :return an array containing detectedAreas with projected coordinates
        """
        # Translate the coordinates to the coordinates in the non-rotated image
        for area in rawPositions:
            area.projectArea(rotatedCenter, self.imageCenter)

        return rawPositions


    def HELPER_standardizeCounterClockwiseDetectedArea(self, detectedAreas, rotatedCenter, angle):
        """
        Translate the coordinates of counter clockwise rotated detectedAreas to the original image coordinates
            :param detectedAreas: a list of detectedArea objects
            :param rotatedCenter: the rotated image's center's point object
            :param angle: the angle by which the image was rotated counter clockwise by
            :return a list of translated detectedAreas for the nonrotated image
        """

        AllObjects = []    

        self.HELPER_rotateDetectedAreaClockwise(detectedAreas, rotatedCenter, angle)
        self.HELPER_projectDetectedArea(detectedAreas, rotatedCenter)
        AllObjects.append(detectedAreas)

        return AllObjects
        

    def HELPER_mergeDetectedObjs(self, list):
        """
        Given a list of any number of arrays of (detected objects, rotatedCenter), scan through all of them and if find two duplicates (similar detected objects with similar 
        sizes and positions), merge them and put all the unique detected object in an array. 
            :param list: a list of arrays
            :return an array that contains all the unique detected objects.
        """
        array0 = list[0]
        others = []
        for i in range(1,len(list)):
            others.append(list[i])


        # checking duplicates in the first array
        i = 0
        while i < len(array0) - 1:
            j = i + 1
            while j < len(array0):
                if array0[i].similarSize(array0[j], self.HARDCODED_similarSizeScale) and array0[i].overlap(array0[j]):
                        array0[i].merge(array0[j])
                        array0.pop(j)
                        j = j - 1
                j = j + 1
            i = i + 1

        # check for duplicates in the other arrays. Merge them if they are, append them if they arent
        for array in others:
            for area in array:
                matched = False
                # Scan through the array to find matches
                for area0 in array0:
                    if area0.similarSize(area, self.HARDCODED_similarSizeScale) and area0.overlap(area):
                        area0.merge(area)
                        matched = True
                        break
                
                # if didnt find any matches then append it onto array0
                if matched == False:
                    array0.append(area)    
        
        return array0


    def HELPER_findEyesCounterClockwiseAngle(self, angle, scaleFactor, minNeighbors):
        """
        Find the (non-paired) eyes in the counter clockwise rotated image. Merge the duplicates and return them in an array
            :param angle: the angle by which the image is rotated by counter clockwise
            :param scaleFactor: detectMultiscale parameter
            :param minNeighbors: detectMultiscale parameter
            :return an array of detected eyes in the counter clockwise rotated image
        """
        (detectedEyes, rotatedCenter) = self.HELPER_runHaarDetectionCounterClockwiseAngle(self.haarcascade_eye, self.HARDCODED_eyeMinDimensions, self.HARDCODED_eyeMaxDimensions, angle, scaleFactor, minNeighbors)
        eyes = self.HELPER_mergeDetectedObjs(self.HELPER_standardizeCounterClockwiseDetectedArea(detectedEyes, rotatedCenter, angle))
           
        return eyes



    def findPairsOfEyesCounterClockwiseAngle(self, angle, scaleFactor, minNeighbors):
        """
        Find the pairs of eyes in the counter clockwise rotated image. Return them in an array containing 2-tuple with the 
        first element being the left-most eye and the second element the right-most eye.
            :param angle: the angle by which the image is rotated by counter clockwise
            :param scaleFactor: detectMultiscale parameter
            :param minNeighbors: detectMultiscale parameter
            :return an array of pairs of eyes 
        """
        
        eyes = self.HELPER_findEyesCounterClockwiseAngle(angle, scaleFactor, minNeighbors)

        pairOfEyes = []
        
        for i in range(len(eyes) - 1):
            for j in range(i,len(eyes)):
                if eyes[i].similarSize(eyes[j], self.HARDCODED_similarSizeScale):
                    dist = eyes[i].center.distTo(eyes[j].center)
                    averageRadius = (eyes[1].radius + eyes[j].radius)/2
                    if dist < self.HARDCODED_pairOfEyesDistanceRange[1] * averageRadius and dist > self.HARDCODED_pairOfEyesDistanceRange[0] * averageRadius:
                        # Let the left most eye be the first eye. This is for calculating relative angle of the face in findFacesUsingPairOfEyes method
                        if eyes[i].center.x < eyes[j].center.x:
                            pairOfEyes.append((eyes[i], eyes[j]))
                        else:
                            pairOfEyes.append((eyes[j], eyes[i]))

        return pairOfEyes


    def findPairsOfEyesCounterClockwiseMultipleAngles(self, angles, scaleFactor, minNeighbors):
        """
        Find the pairs of eyes in the counter clockwise rotated images. Return them in an array containing 2-tuple with the 
        first element being the left-most eye and the second element the right-most eye.
            :param angles: the angles by which the image is rotated by counter clockwise
            :param scaleFactor: detectMultiscale parameter
            :param minNeighbors: detectMultiscale parameter
            :return an array of pairs of eyes 
        """
        eyes = []
        for angle in angles:
            eyes.append(self.HELPER_findEyesCounterClockwiseAngle(angle, scaleFactor, minNeighbors))

        eyes = self.HELPER_mergeDetectedObjs(eyes)

        pairOfEyes = []
        
        for i in range(len(eyes) - 1):
            for j in range(i,len(eyes)):
                if eyes[i].similarSize(eyes[j], self.HARDCODED_similarSizeScale):
                    if eyes[i].appropriateDistanceTo(eyes[j], self.HARDCODED_pairOfEyesDistanceRange[0], self.HARDCODED_pairOfEyesDistanceRange[1]):
                        # Let the left most eye be the first eye. This is for calculating relative angle of the face in findFacesUsingPairOfEyes method
                        if eyes[i].center.x < eyes[j].center.x:
                            pairOfEyes.append((eyes[i], eyes[j]))
                        else:
                            pairOfEyes.append((eyes[j], eyes[i]))

        return pairOfEyes

 
    def DEBUG_findFacesUsingPairOfEyes(self, pairOfEyes, scaleFactor, minNeighbors):
        """
        Using given pairs of eyes, for each pair find the angle the face is leaning, crop the area the face could be out and run haarDetection on that area.
        Return an array of all detectedAreas encapsulating faces.
            :param pairOfEyes: 2-tuple (left eye, right eye)
            :param scaleFactor: parameter for detectMultiScale
            :param minNeighbors: parameter for detectedMultiScale
            return array of all detected faces
        """

        # For right now, let face width be 6 average radius, and height be 10 average radius with 5 average radiuses from 2 eyes center to the top 
        # 5 average radiuses from 2 eyes center to the chin

        debugArrayFaces = []
        debugPotentialFaceNumber = 1
        for pair in pairOfEyes:
            
            leftEye = pair[0]
            rightEye = pair[1]
            eyeAverageRadius = (leftEye.radius + rightEye.radius)/2
            # halfFaceDimensions store the distance from faceOrigin to the left border, to the upper border, and to the lower border
            # faceMinRadius is the min radius of the rectangle encapsulating the detected Face
            halfFaceDimensions = (eyeAverageRadius * 4, eyeAverageRadius * 5, eyeAverageRadius * 5)
            faceMinRadius = eyeAverageRadius * 3

            # relative angle is the relative angle of the right eye to the left eye, but I limit the ranges from 270 -> 0 -> 90 degree because faces usually aren't up side down, 
            # still this function takes care of that case also (scan it upside down when find no face)
            relativeCounterClockwiseAngle = (leftEye.center.relativeCounterClockwiseAngle(rightEye.center) + 90) % 180 - 90
            originalImageCenter = Point(self.grayImage.shape[1]/2, self.grayImage.shape[0]/2)
            faceOrigin = Point((leftEye.center.x + rightEye.center.x)/2,(leftEye.center.y + rightEye.center.y)/2)
           
            # Rotate the face origin point and the image such that the face is straightened up
            rotatedGrayImage = rotateClockwise(self.grayImage, relativeCounterClockwiseAngle)
            rotatedImageCenter = Point(rotatedGrayImage.shape[1]/2, rotatedGrayImage.shape[0]/2)
            rotatedFaceOrigin = faceOrigin.rotatePointClockwise(originalImageCenter, relativeCounterClockwiseAngle)
            rotatedFaceOrigin = rotatedFaceOrigin.projectPoint(originalImageCenter, rotatedImageCenter)
            
            
            
            # DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- 
            debugOriginalImage = rotateClockwise(self.image, 0)
            leftEye.draw(debugOriginalImage, (255,0,255), 2)
            rightEye.draw(debugOriginalImage, (0,255,255), 2)
            cv.circle(debugOriginalImage, faceOrigin.exportCoordinates(), 0, (0, 255, 0), 20)
            debugRotatedOriginalImage = rotateClockwise(debugOriginalImage, relativeCounterClockwiseAngle)
            cv.circle(debugRotatedOriginalImage, rotatedFaceOrigin.exportCoordinates(), 0, (255, 255, 255), 12)
            # DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- DEBUG step --- 
            


            # Crop the potential face out
            cropRangeMinY = int(max(0, rotatedFaceOrigin.y - halfFaceDimensions[1]))
            cropRangeMaxY = int(min(rotatedGrayImage.shape[0], rotatedFaceOrigin.y + halfFaceDimensions[2]))

            cropRangeMinX = int(max(0, rotatedFaceOrigin.x - halfFaceDimensions[0]))
            cropRangeMaxX = int(min(rotatedGrayImage.shape[1], rotatedFaceOrigin.x + halfFaceDimensions[0]))

            try:
                rotatedCrop = np.zeros((cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX), dtype='uint8')
                rotatedCrop[0: rotatedCrop.shape[0], 0: rotatedCrop.shape[1]] = \
                    rotatedGrayImage[cropRangeMinY:cropRangeMaxY, cropRangeMinX:cropRangeMaxX]
             
            except:
                rotatedCrop = np.zeros((cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX, 3), dtype ='uint8')
                rotatedCrop[0: rotatedCrop.shape[0], 0: rotatedCrop.shape[1]] = \
                    rotatedGrayImage[cropRangeMinY:cropRangeMaxY, cropRangeMinX:cropRangeMaxX]
   
            try:
                debugRotatedCrop = np.zeros((cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX), dtype ='uint8')
                debugRotatedCrop[0: debugRotatedCrop.shape[0], 0: debugRotatedCrop.shape[1]] = debugRotatedOriginalImage[cropRangeMinY:cropRangeMaxY, cropRangeMinX:cropRangeMaxX]
            except:
                debugRotatedCrop = np.zeros((cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX, 3), dtype ='uint8')
                debugRotatedCrop[0: debugRotatedCrop.shape[0], 0: debugRotatedCrop.shape[1]] = debugRotatedOriginalImage[cropRangeMinY:cropRangeMaxY, cropRangeMinX:cropRangeMaxX]

        
            rotatedFaceCenter = Point((cropRangeMinX + cropRangeMaxX)/2, (cropRangeMinY + cropRangeMaxY)/2)
            croppedCenter = Point(rotatedCrop.shape[1]/2, rotatedCrop.shape[0]/2)


            # find the face in the cropped Area
            detectedFaces = self.haarcascade_face.detectMultiScale(rotatedCrop, scaleFactor, minNeighbors, minSize = self.HARDCODED_faceMinDimensions, maxSize = self.HARDCODED_faceMaxDimensions)


            # DEBUGING AREA
            debugRotatedCrop = cv.circle(debugRotatedCrop, croppedCenter.exportCoordinates(), 0, (255, 255, 255), 6)

            # DEBUGING AREA


            boolUpSideDown = False

            if len(detectedFaces) == 0:
                # Scan the image upside down in case of upside down faces
                rotatedCrop = rotateClockwise(rotatedCrop, 180)
                detectedFaces = self.haarcascade_face.detectMultiScale(rotatedCrop, scaleFactor, minNeighbors, minSize = self.HARDCODED_faceMinDimensions, maxSize = self.HARDCODED_faceMaxDimensions)
                if len(detectedFaces) == 0:
                    continue
                boolUpSideDown = True


            # merge smaller faces to the biggest face
            biggestFace = detectedFaces[0]
            for i in range(1, len(detectedFaces)):
                # Doesn't have to be too complicated, if one dimension is larger the other is 99% of the time larger as well
                if detectedFaces[i][2] > biggestFace[2]:
                    biggestFace = detectedFaces[i]


            # if face's radius is too small then its not a face (face right now is a 4-tuple (x, y, w, h))
            if biggestFace[2] ** 2 + biggestFace[3] ** 2 < (faceMinRadius * 2) ** 2:

                if boolUpSideDown:
                    continue

                # Scan the image upside down in case of upside down faces
                rotatedCrop = rotateClockwise(rotatedCrop, 180)
                detectedFaces = self.haarcascade_face.detectMultiScale(rotatedCrop, scaleFactor, minNeighbors, minSize = self.HARDCODED_faceMinDimensions, maxSize = self.HARDCODED_faceMaxDimensions)
                if len(detectedFaces) == 0:
                    continue
                boolUpSideDown = True

                
                # merge smaller faces to the biggest face
                biggestFace = detectedFaces[0]
                for i in range(1, len(detectedFaces)):
                    # if one dimension is larger the other is 99% of the time larger as well
                    if detectedFaces[i][2] > biggestFace[2]:
                        biggestFace = detectedFaces[i]

                if biggestFace[2] ** 2 + biggestFace[3] ** 2 < (faceMinRadius * 2) ** 2:
                    continue
            
            # if the face found was upside down, then face angle = eye's relative angle + 180
            if boolUpSideDown:
                counterClockwiseFaceAngle = relativeCounterClockwiseAngle + 180
            # if not then = eye's relative angle
            else:
                counterClockwiseFaceAngle = relativeCounterClockwiseAngle

            biggestFace = DetectedFace((biggestFace[0],biggestFace[1]), (biggestFace[2], biggestFace[3]), counterClockwiseFaceAngle)

            
            if boolUpSideDown:
                biggestFace.rotateAreaClockwise(croppedCenter, 180)

            # Convert biggestFace coordinates from being in the cropped image to the original image
            biggestFace.draw(debugRotatedCrop, (0, 0, 255), 2)
            cv.imshow("Potential face " + str(debugPotentialFaceNumber), resizeMinTo(debugRotatedCrop, 250))

            debugArrayFaces.append((biggestFace.copy(), rotatedFaceCenter, croppedCenter, rotatedImageCenter, relativeCounterClockwiseAngle, originalImageCenter)) 
                
            
            debugPotentialFaceNumber = debugPotentialFaceNumber + 1


        return debugArrayFaces


    def findFacesUsingPairOfEyes(self, pairOfEyes, scaleFactor, minNeighbors):
        """
        Using given pairs of eyes, for each pair of similar-size eyes, find the angle the face is leaning, crop the area the face could be out and run haarDetection on that area.
        Return an array of all detectedAreas encapsulating faces.
            :param pairOfEyes: 2-tuple (left eye, right eye)
            :param scaleFactor: parameter for detectMultiScale
            :param minNeighbors: parameter for detectedMultiScale
            return array of all detected faces
        """

        faces = [] 

        # For right now, let face width be 6 average radius, and height be 10 average radius with 5 average radiuses from 2 eyes center to the top 
        # 5 average radiuses from 2 eyes center to the chin. I'm going to make the height ratio 3 top:5 bottom once i added mouth/nose detection for orientation
        
        for pair in pairOfEyes:
            
            leftEye = pair[0]
            rightEye = pair[1]
            eyeAverageRadius = (leftEye.radius + rightEye.radius)/2
            # halfFaceDimensions store the distance from faceOrigin to the left border, to the upper border, and to the lower border
            # faceMinRadius is the min radius of the rectangle encapsulating the detected Face
            halfFaceDimensions = (eyeAverageRadius * 4, eyeAverageRadius * 5, eyeAverageRadius * 5)
            faceMinRadius = eyeAverageRadius * 3

            # relative angle is the relative angle of the right eye to the left eye, but I limit the ranges from 270 -> 0 -> 90 degree because faces usually aren't up side down, 
            # still this function takes care of that case also (scan it upside down when find no face)
            relativeCounterClockwiseAngle = (leftEye.center.relativeCounterClockwiseAngle(rightEye.center) + 90) % 180 - 90
            originalImageCenter = Point(self.grayImage.shape[1]/2, self.grayImage.shape[0]/2)
            faceOrigin = Point((leftEye.center.x + rightEye.center.x)/2,(leftEye.center.y + rightEye.center.y)/2)
           
            # Rotate the face origin point and the image such that the face is straightened up
            rotatedGrayImage = rotateClockwise(self.grayImage, relativeCounterClockwiseAngle)
            rotatedImageCenter = Point(rotatedGrayImage.shape[1]/2, rotatedGrayImage.shape[0]/2)
            rotatedFaceOrigin = faceOrigin.rotatePointClockwise(originalImageCenter, relativeCounterClockwiseAngle).projectPoint(originalImageCenter, rotatedImageCenter)

            # Crop the potential face out
            cropRangeMinY = int(max(0, rotatedFaceOrigin.y - halfFaceDimensions[1]))
            cropRangeMaxY = int(min(rotatedGrayImage.shape[0], rotatedFaceOrigin.y + halfFaceDimensions[2]))

            cropRangeMinX = int(max(0, rotatedFaceOrigin.x - halfFaceDimensions[0]))
            cropRangeMaxX = int(min(rotatedGrayImage.shape[1], rotatedFaceOrigin.x + halfFaceDimensions[0]))

            try:
                rotatedCrop = np.zeros((cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX), dtype='uint8')
                rotatedCrop[0: rotatedCrop.shape[0], 0: rotatedCrop.shape[1]] = \
                    rotatedGrayImage[cropRangeMinY:cropRangeMaxY, cropRangeMinX:cropRangeMaxX]
             
            except:
                rotatedCrop = np.zeros((cropRangeMaxY - cropRangeMinY, cropRangeMaxX - cropRangeMinX, 3), dtype ='uint8')
                rotatedCrop[0: rotatedCrop.shape[0], 0: rotatedCrop.shape[1]] = \
                    rotatedGrayImage[cropRangeMinY:cropRangeMaxY, cropRangeMinX:cropRangeMaxX]
        
            rotatedFaceCenter = Point((cropRangeMinX + cropRangeMaxX)/2, (cropRangeMinY + cropRangeMaxY)/2)
            croppedCenter = Point(rotatedCrop.shape[1]/2, rotatedCrop.shape[0]/2)


            # find the face in the cropped Area
            detectedFaces = self.haarcascade_face.detectMultiScale(rotatedCrop, scaleFactor, minNeighbors, minSize = self.HARDCODED_faceMinDimensions, maxSize = self.HARDCODED_faceMaxDimensions)
            boolUpSideDown = False


            # if found no faces right side up
            if len(detectedFaces) == 0:
                # Scan the image upside down
                rotatedCrop = rotateClockwise(rotatedCrop, 180)
                detectedFaces = self.haarcascade_face.detectMultiScale(rotatedCrop, scaleFactor, minNeighbors, minSize = self.HARDCODED_faceMinDimensions, maxSize = self.HARDCODED_faceMaxDimensions)
                if len(detectedFaces) == 0:
                    continue
                boolUpSideDown = True


            # merge smaller faces to the biggest face
            biggestFace = detectedFaces[0]
            for i in range(1, len(detectedFaces)):
                # Doesn't have to be too complicated, if one dimension is larger the other is 99% of the time larger as well
                if detectedFaces[i][2] > biggestFace[2]:
                    biggestFace = detectedFaces[i]


            # if face's radius is too small then its not a face (face right now is a 4-tuple (x, y, w, h))
            if biggestFace[2] ** 2 + biggestFace[3] ** 2 < (faceMinRadius * 2) ** 2:

                if boolUpSideDown:
                    continue

                # Scan the image upside down in case of upside down faces
                rotatedCrop = rotateClockwise(rotatedCrop, 180)
                detectedFaces = self.haarcascade_face.detectMultiScale(rotatedCrop, scaleFactor, minNeighbors, minSize = self.HARDCODED_faceMinDimensions, maxSize = self.HARDCODED_faceMaxDimensions)
                if len(detectedFaces) == 0:
                    continue
                boolUpSideDown = True

                
                # merge smaller faces to the biggest face
                biggestFace = detectedFaces[0]
                for i in range(1, len(detectedFaces)):
                    # if one dimension is larger the other is 99% of the time larger as well
                    if detectedFaces[i][2] > biggestFace[2]:
                        biggestFace = detectedFaces[i]

                # if face's radius is too small then its not a face (face right now is a 4-tuple (x, y, w, h))
                if biggestFace[2] ** 2 + biggestFace[3] ** 2 < (faceMinRadius * 2) ** 2:
                    continue



            biggestFace = DetectedFace((biggestFace[0],biggestFace[1]), (biggestFace[2], biggestFace[3]))
            
            
            


            if boolUpSideDown:
                biggestFace.rotateAreaClockwise(croppedCenter, 180)

            # Convert biggestFace coordinates from being in the cropped image to the original image
            biggestFace.projectArea(croppedCenter, rotatedFaceCenter)
            biggestFace.rotateAreaCounterClockwise(rotatedImageCenter, relativeCounterClockwiseAngle)
            biggestFace.projectArea(rotatedImageCenter, originalImageCenter)


            
            # if the face found was upside down, then face angle = eye's relative angle + 180, the eyes are swapped
            if boolUpSideDown:
                counterClockwiseFaceAngle = (relativeCounterClockwiseAngle + 180) % 360
                biggestFace.counterClockwiseAngle = counterClockwiseFaceAngle
                biggestFace.leftEye = rightEye
                biggestFace.rightEye = leftEye
            # if not then = eye's relative angle, the eyes are in the same order
            else:
                counterClockwiseFaceAngle = relativeCounterClockwiseAngle
                biggestFace.counterClockwiseAngle = counterClockwiseFaceAngle
                biggestFace.leftEye = leftEye
                biggestFace.rightEye = rightEye
            

            faces.append(biggestFace)

        return faces


    def findFacesCounterClockwiseAngle(self, angle, scaleFactor, minNeighbors):
        """
        Find faces in the image and return them as detectedArea objects in an array
            :param angle: counter clockwise angle by which the image is rotated
            :param scaleFactor: paramter for detectMultiScale
            :param minNeighbors: parameter for detectMultiScale
            :return an array of faces as detectedFace objects
        """
        return self.findFacesUsingPairOfEyes(self.findPairsOfEyesCounterClockwiseAngle(angle, scaleFactor, minNeighbors), scaleFactor, minNeighbors)


    def findFacesCounterClockwiseMultipleAngles(self, angles, scaleFactor, minNeighbors):
        """
        Find faces in the image and return them as detectedArea objects in an array
            :param angles: counter clockwise angles by which the image is rotated
            :param scaleFactor: paramter for detectMultiScale
            :param minNeighbors: parameter for detectMultiScale
            :return an array of faces as detectedFace objects
        """
        return self.findFacesUsingPairOfEyes(self.findPairsOfEyesCounterClockwiseMultipleAngles(angles, scaleFactor, minNeighbors), scaleFactor, minNeighbors)

                














from . import Point
import cv2 as cv

"""
NOTE:
    Only faces have their own class DetectedFace because I want to keep track of the angle of the face using the relative angle between the two eyes (and potentially an eye and a nose or nose and mouth, etc)
    The other facial features such as eyes, noses, mouths dont need to have their own class and can just use DetectedArea to store their locations
"""


class DetectedArea:
    """
    Object that stores the center of a rectangle encapsulates a detected object returned by cv.detectMultiscale(...), and the dimensions of the rectangle.
    OpenCV haarcascade detection returns 4-tuples (x,y,w,h) of rectangle marking down where the detected objects are. These information will be translated 
    to DetectedArea objects.
    """
    def __init__(self, upperLeftPoint = (0,0), dimensions = (0,0)):
        """
        Construct a DetectedArea obj.
            :param upperLeftPoint (x,y): the 2-tuple coordinates of the upper left point of the rectangle encapsulates detected objects.
            :param dimension (w,h): the dimension of the box encapsulates the detected object.
            :param angle: the angle of the image when the detected object was found and returned by openCV.
        """
        self.dimensions = dimensions
        self.upperLeft = Point(upperLeftPoint[0], upperLeftPoint[1])
        self.upperRight = Point(upperLeftPoint[0] + dimensions[0], upperLeftPoint[1])
        self.lowerRight = Point(upperLeftPoint[0] + dimensions[0], upperLeftPoint[1] + dimensions[1])
        self.lowerLeft = Point(upperLeftPoint[0], upperLeftPoint[1] + dimensions[1])
        self.center = Point(upperLeftPoint[0] + dimensions[0]/2, upperLeftPoint[1] + dimensions[1]/2)
        self.radius = self.center.distTo(self.upperLeft)
    
    def copy(self):
        """
        Return a deep copy of the detectedArea caller
            :return a deep copy of itself
        """
        copyArea = DetectedArea()
        copyArea.dimensions = (self.dimensions[0], self.dimensions[1])
        copyArea.upperLeft = self.upperLeft.copy()
        copyArea.upperRight = self.upperRight.copy()
        copyArea.lowerLeft = self.lowerLeft.copy()
        copyArea.lowerRight = self.lowerRight.copy()
        copyArea.center = self.center.copy()
        return copyArea
    
    def rotateAreaCounterClockwise(self, origin, angle):
        """
        Rotate the detectedArea object counter-clockwise around an origin by a given degree.
            :param origin: the point where detectedArea is going to rotate around
            :param angle: the angle the detectedArea is going to rotate by
        """
        self.upperLeft = self.upperLeft.rotatePointCounterClockwise(origin, angle)
        self.upperRight = self.upperRight.rotatePointCounterClockwise(origin, angle)
        self.lowerLeft = self.lowerLeft.rotatePointCounterClockwise(origin, angle)
        self.lowerRight = self.lowerRight.rotatePointCounterClockwise(origin, angle)
        self.center = self.center.rotatePointCounterClockwise(origin, angle)



    def rotateAreaClockwise(self, origin, angle):
        """
        Rotate the detectedArea object clockwise around an origin by a given degree.
            :param origin: the point where detectedArea is going to rotate around
            :param angle: the angle the detectedArea is going to rotate by
        """
        self.upperLeft = self.upperLeft.rotatePointClockwise(origin, angle)
        self.upperRight = self.upperRight.rotatePointClockwise(origin, angle)
        self.lowerLeft = self.lowerLeft.rotatePointClockwise(origin, angle)
        self.lowerRight = self.lowerRight.rotatePointClockwise(origin, angle)
        self.center = self.center.rotatePointClockwise(origin, angle)
        
    def projectArea(self, oldOrigin, newOrigin):
        """
        Project the rectangle such that its new position relative to newOrigin is 
        the same as its current relative position to oldOrigin.
            :param oldOrigin: the old origin point
            :param newOrigin: the projected old Origin
        """        
        self.upperLeft = self.upperLeft.projectPoint(oldOrigin, newOrigin)
        self.upperRight = self.upperRight.projectPoint(oldOrigin, newOrigin)
        self.lowerLeft = self.lowerLeft.projectPoint(oldOrigin, newOrigin)
        self.lowerRight = self.lowerRight.projectPoint(oldOrigin, newOrigin)
        self.center = self.center.projectPoint(oldOrigin, newOrigin)

    def overlap(self, otherArea):
        """
        Check for overlap between two detectedAreas. (NOT FINALIZED CONDITION PARAMETERS)
            :param otherArea: the DetectedArea that we are going to check for overlap
            :return True/False
        """
        # print("DetectedArea overlap: NOT FINALIZED CONDITION PARAMETERS")
        distance = self.center.distTo(otherArea.center)
        if distance < (self.center.distTo(self.upperLeft) + otherArea.center.distTo(otherArea.upperLeft))/4:
            return True
        return False
    
    def similarSize(self, otherArea, scale):
        """
        Compare the size with another area to see if the two DetectedAreas are similar. (NOT FINALIZED CONDITION PARAMETERS)
            :param otherArea: the DetectedArea that we are comparing to
            :param scale: the smallest possible size the smaller area can be compared to the bigger area to be considered "similarSize"
            :return True/False
        """
        thisSize = self.center.distTo(self.upperLeft)
        otherSize = otherArea.center.distTo(otherArea.upperLeft)

        minSize = min(thisSize, otherSize)
        maxSize = max(thisSize, otherSize)

        if minSize > maxSize * scale:
            return True
        return False
    
    def appropriateDistanceTo(self, otherArea, minAverageRadiusDistScale, maxAverageRadiusDistScale):
        """
        Decide whether the two detectedAreas distance to each other is appropriate.
            :param otherArea: the other detectedArea object
            :param minAverageRadiusDistScale: the mininum multiplier for the average radius between two detectedAreas 
            :param maxAverageRadiusDistScale: the maximum multiplier for the average radius between two detectedAreas 
            :return True if the distance between two objects is between average radius * minimum multiplier and average radius * maximum multiplier, False otherwise
        """
        dist = self.center.distTo(otherArea.center)
        averageRadius = (self.radius + otherArea.radius)/2
        if dist < maxAverageRadiusDistScale * averageRadius and dist > minAverageRadiusDistScale * averageRadius:
            return True
        return False

    def merge(self, otherArea):
        """
        Merge another area into itself to make a more accurate area. This does not delete the other area object. (NOT FINALIZED CONDITION PARAMETERS)
        For right now, merge will trust the user to call appropriately since all it does is make the caller object becomes the smaller object between the twos.
            :param otherArea: the DetectedArea that we are merging with
        """
        if self.center.distTo(self.upperLeft) > otherArea.center.distTo(otherArea.upperLeft):
            self = otherArea
        
        

    def draw(self, canvas, color, thickness):
        """
        Draw the shape of the detectedArea on top of the given image with specified color and thickness
            :param canvas: the given image to be drawn over
            :param color: the BGR-color specified for the detectedArea object
            :param thickness: the specified thickness of the detectedArea object
        """
        upperLeft = self.upperLeft.exportCoordinates()
        upperRight = self.upperRight.exportCoordinates()
        lowerLeft = self.lowerLeft.exportCoordinates()
        lowerRight = self.lowerRight.exportCoordinates()
        canvas = cv.line(canvas, upperLeft, upperRight, color, thickness = thickness)
        canvas = cv.line(canvas, upperRight, lowerRight, color, thickness = thickness)
        canvas = cv.line(canvas, lowerRight, lowerLeft, color, thickness = thickness)
        canvas = cv.line(canvas, lowerLeft, upperLeft, color, thickness = thickness)
        


class DetectedFace(DetectedArea):
    def __init__(self, upperLeftPoint = (0,0), dimensions = (0,0), angle = None):
        """
        Create an detectedFace object that is a detectedArea object but with angle property.
            :param upperLeftPoint (x,y): the 2-tuple coordinates of the upper left point of the rectangle encapsulates detected objects.
            :param dimension (w,h): the dimension of the box encapsulates the detected object.
            :param angle: the angle of the image when the detected object was found and returned by openCV.
        """
        super().__init__(upperLeftPoint=upperLeftPoint, dimensions=dimensions)
        # face's angle is the counter clockwise angle of the right eye to the left eye
        self.counterClockwiseAngle = angle
        self.leftEye = None
        self.rightEye = None
        
    
    def copy(self):
        """
        Return a deep copy of the detectedArea caller
            :return a deep copy of itself
        """
        copyArea = DetectedFace()
        copyArea.dimensions = (self.dimensions[0], self.dimensions[1])
        copyArea.upperLeft = self.upperLeft.copy()
        copyArea.upperRight = self.upperRight.copy()
        copyArea.lowerLeft = self.lowerLeft.copy()
        copyArea.lowerRight = self.lowerRight.copy()
        copyArea.center = self.center.copy()
        copyArea.counterClockwiseAngle = self.counterClockwiseAngle
        copyArea.leftEye = self.leftEye
        copyArea.rightEye = self.rightEye
        return copyArea